/*	fkt_print.c

	fkt_print --	program to analyse and print trace files produced by
					fkt_record using fkt (fast kernel tracing) in the kernel.

	Copyright (C) 2000, 2001  Robert D. Russell -- rdr@unh.edu

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/

/*	As of 19-Oct-00,
	unshifted codes (for traps/syscalls/irqs) MUST be less than
		FKT_UNSHIFTED_LIMIT_CODE.
	singleton codes (i.e., those that are NOT paired in entry-exit pairs)
		MUST be greater than or equal to FKT_UNSHIFTED_LIMIT_CODE
		and less than FKT_RET_FROM_SYS_CALL_CODE.
	function entry and exit codes MUST be greater than
		FKT_RET_FROM_SYS_CALL_CODE == 0x2ff.
	function entry codes MUST have the FKT_GENERIC_EXIT_OFFSET bit set.
	function exit  codes MUST NOT have the FKT_GENERIC_EXIT_OFFSET bit set.
	function exit codes = function entry codes + FKT_GENERIC_EXIT_OFFSET.
*/

/*****	15-Oct-00	rdr		modified for 2.4.0-test9 *****/
/*	also commented out use of FKT_TCP_RECVMSG_SCHEDI_CODE */

/*	Usage: fkt_print [-f input-file] [-p] [-c] [-d] [-a]

			if -f is not given, look for TRACE_FILE environment variable, and
				if that is not given, read from "trace_file".

			if -p is given, printing is forced for all trace items
				if that is not given, printing starts with first item belong to
				task having kidpid

			if -c is given, time printed is cummulative time
				if that is not given, time is since previous item printed

			if -d is given, turn on debug printout to stdout for stack nesting
				if that is not given, no debug printout is printed

			if -a is given, accumulate statistics for all trace items
				if that is not given, accumulate stats only between the first
				and second times() system call by the task having kidpid
*/

#define _POSIX_C_SOURCE 199309L

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <linux/unistd.h>
#include <linux/fkt.h>


#define WHITE_SPACE " \t\n\f\v\r"
#define OPTIONS ":f:pcda"
#define DEFAULT_TRACE_FILE "trace_file"

#define HISTO_SIZE	128
#define MAX_HISTO_BAR_LENGTH	29
#define HISTO_CHAR	'*'

static int print_delta_time = 1;	/* 1 to print delta, 0 for cummulative */
static FILE *stdbug = NULL;			/* fd for debug file (/dev/null normally) */

/*	use 64-bit arithmetic for all times */
typedef unsigned long long int	u_64;


/*
Document regarding printing cycles for each function 

MAX_FUNS is to be set to the number of functions that contain probes.
MAX_EACH_FUN_RECUR is the number of times that the same function
could be calling itself recursively.
MAX_UNSHIFTED_FUN_RECUR - is used to store the cycles for all codes
less than 0x200. that is, these codes are all entered in a single array
and when a return from system call is seen, then the last element
from the array is popped out and subtracted from the cycle at the
return from system call .

"fun_cycles_time" is the structure that holds a counter ( for the number
of times the "entry" to the function is seen), the code ( for the
"entry" probes to the functions), and the total cycles for each function
is maintained in the field total.
cycles_array is used to store the cycles when the same function is called
recursively.

Exit codes for each function are 0x100 more than the entry probe codes.
Therefore, the way that the time spent on executing a function is calculated
is as follows.  When a function is entered for the first time, store
it in the array "v", with its entry code, and cycles in the cycles_array.
when the exit code is seen for that function, index into the "v" array
and subtract the entry cycles from the exit cycle.  The difference is stored
in the field "total" cumulatively.  If before seeing an exit, another entry
for the same function is seen, no problem, this can also be stored
in the cycles_array..and popped out appropriately...
*/

#define MAX_FUNS 200
#define MAX_EACH_FUN_RECUR 10
#define MAX_UNSHIFTED_FUN_RECUR 16
#define MAX_SHIFTED_FUN_RECUR 64

/*	Variables for usage statistics of various pre-sized arrays.
	These always hold the last used value of the indexes.
*/
static unsigned int		actual_max_v2_pos = 0;
static unsigned int		actual_max_v3_pos = 0;
static unsigned int		actual_max_v_pos = 0;
static unsigned int		actual_max_cycles_pos = 0;
static unsigned int		actual_max_v3_function = 0;


/*	One of these items in v[] array for each function called */
struct fun_cycles_item	{
										/* the code identifying this func */
						int				code;

										/* total number of calls to this func */
						int				counter;

										/* total number of cycles in this func*/
						u_64			total;

										/* stack of start times of open calls to
										   this func (0 if not open call) */
						u_64			cycles_array[MAX_EACH_FUN_RECUR];

										/* stack of pid that made open call */
						unsigned int	caller_pid[MAX_EACH_FUN_RECUR];

										/* ith slot is cycles spent in func i
										   when called from this func */
						u_64			called_funs_cycles[MAX_FUNS];

										/* ith slot is number of calls to
										   func i from this func */
						int				called_funs_ctr[MAX_FUNS];
						};

/* The v[] array stores information for unique functions actually called */
static int v_pos = 0;					/* index to next slot to use in v[] */
struct fun_cycles_item v[MAX_FUNS];


/*	arrays to keep track of number of occurrences and time accumulated
	in traps/syscalls/irqs
*/
static u_64			syscall_cycles[FKT_UNSHIFTED_LIMIT_CODE] = {0};
static unsigned int	syscall_count[FKT_UNSHIFTED_LIMIT_CODE] = {0};


/*	v3 stacks keep track of start cycle and code for nested function calls.
	One stack per pid, kept in linked-list headed by v3_stack_head.
	Current stack is pointed to by v3_stack_ptr[lastcpu]. */

/*	Keep kernel function nesting statistics in the v3 arrays.
	Every cycle is accounted for exactly once.

	Each time a kernel function is entered, push (v3_pos++)
	onto the v3_code stack its entry code,
	and onto the v3_start_cycle stack the time it occurred.

	When the corresponding kernel function returns, pop (v3_pos--)
	the v3 stacks, using the difference between the v3_start_cycle and the
	time of the return to compute the time spent in the function.
	Add this time to all the other v3_start_cycle times in this stack
	(thereby removing the cycles spent in this nested function from those
	that called it), and, in update_function, add it to the v3_function_cycles
	array at the same index as the one where this function code is stored in
	the v3_function_code array.  Also increment the v3_function_count array
	slot at the same index. */

/*	Keep track of this pid's open traps/syscalls/irqs in the v2 stacks.

	Each time a trap/syscall/irq is encountered, push (v2_pos++)
	onto the v2_code stack its entry code,
	and onto the v2_start_cycle stack the time it occurred.

	On a return from sys call, pop (v2_pos--) the v2 stacks,
	using the difference between the v2_start_cycle and the time of the
	return to compute the time spent in the trap/syscall/irq.
	This time spent is added to the syscall_cycles array at the position
	indexed by the entry code at the top of the v2_code_stack.
	The syscall_count array at that same position is also incremented.
*/

struct v3_stack_item
			{
							/* the pid to whom this item belongs */
			unsigned int	v3_pid;
							/* the last cpu that worked on this item */
			unsigned int	v3_cpu;
							/* total number of cycles used by this pid */
			u_64			v3_process_cycles;

							/* index to next unused slot on v3 stacks */
			int				v3_pos;
							/* stack of start cycles for open function calls */
			u_64			v3_start_cycle[MAX_SHIFTED_FUN_RECUR];
							/* stack of entry codes for open function calls */
			unsigned int	v3_code[MAX_SHIFTED_FUN_RECUR];
							/* stack of indexes to entry codes in v array */

			int				v_index_array[MAX_FUNS] ;
							/* index to next unused slot on index_array */
			int				v_index_arr_pos;

							/* index to last used slot on v2 stacks */
			int				v2_pos;
							/* stack of start cycle for open trap/syscall/irq */
			u_64			v2_start_cycle[MAX_UNSHIFTED_FUN_RECUR];
							/* stack of entry code for open trap/syscall/irq */
			unsigned int	v2_code[MAX_UNSHIFTED_FUN_RECUR];

									/* link to item for next pid */
			struct v3_stack_item	*v3_next;
			};

static struct v3_stack_item	*v3_stack_head = NULL;
static struct v3_stack_item	*v3_stack_first = NULL;
static struct v3_stack_item	**v3_stack_ptr = NULL;


/*	arrays to keep track of number of occurrences and time accumulated */
static u_64			v3_syscall_cycles[FKT_UNSHIFTED_LIMIT_CODE] = {0};
static unsigned int v3_syscall_count[FKT_UNSHIFTED_LIMIT_CODE] = {0};
static u_64			v3_function_cycles[MAX_FUNS] = {0};
static unsigned int v3_function_count[MAX_FUNS] = {0};
static unsigned int v3_function_code[MAX_FUNS] = {0};

u_64 start_time_glob, end_time_glob;
int a, b;


struct pid_cycles_item	{
						unsigned int			pid;
						u_64					total_cycles;
						struct pid_cycles_item	*link;
						};

struct code_list_item	{
						unsigned int			code;
						char					*name;
						struct code_list_item	*link;
						};


struct code_name	{
					unsigned int		code;
					char				*name;
					};

struct code_name	code_table[] =
			{
			{FKT_SETUP_CODE,				"fkt_setup"},
			{FKT_KEYCHANGE_CODE,			"fkt_keychange"},
			{FKT_RESET_CODE,				"fkt_reset"},
			{FKT_CALIBRATE0_CODE,			"fkt_calibrate0"},
			{FKT_CALIBRATE1_CODE,			"fkt_calibrate1"},
			{FKT_CALIBRATE2_CODE,			"fkt_calibrate2"},
			{FKT_RET_FROM_SYS_CALL_CODE,	"ret_from_sys_call"},

//			{FKT_EXT2_WRITE_ENTRY_CODE,		"ext2_file_write_entry"},
//			{FKT_EXT2_WRITE_EXIT_CODE,		"ext2_file_write_exit"},
			{FKT_LL_RW_BLOCK_ENTRY_CODE,	"ll_rw_block_entry"},
			{FKT_LL_RW_BLOCK_EXIT_CODE,		"ll_rw_block_exit"},
			{FKT_EXT2_GETBLK_ENTRY_CODE,	"ext2_getblk_entry"},
			{FKT_EXT2_GETBLK_EXIT_CODE,		"ext2_getblk_exit"},
//			{FKT_INODE_GETBLK_ENTRY_CODE,	"inode_getblk_entry"},
//			{FKT_INODE_GETBLK_EXIT_CODE,	"inode_getblk_exit"},
//			{FKT_BLOCK_GETBLK_ENTRY_CODE,	"block_getblk_entry"},
//			{FKT_BLOCK_GETBLK_EXIT_CODE,	"block_getblk_exit"},
			{FKT_SWITCH_TO_CODE,			"switch_to"},

			/* sadhna start */

			/* socket interface routines */
			{FKT_SOCKET_ENTRY_CODE,			"socket_entry"},
			{FKT_SOCKET_EXIT_CODE,			"socket_exit"},
			{FKT_BIND_ENTRY_CODE,			"bind_entry"},
			{FKT_BIND_EXIT_CODE,			"bind_exit"},
			{FKT_LISTEN_ENTRY_CODE,			"listen_entry"},
			{FKT_LISTEN_EXIT_CODE,			"listen_exit"},
			{FKT_ACCEPT_ENTRY_CODE,			"accept_entry"},
			{FKT_ACCEPT_EXIT_CODE,			"accept_exit"},
			{FKT_CONNECT_ENTRY_CODE,		"connect_entry"},
			{FKT_CONNECT_EXIT_CODE,			"connect_exit"},
			{FKT_SEND_ENTRY_CODE,			"send_entry"},
			{FKT_SEND_EXIT_CODE,			"send_exit"},
			{FKT_SENDTO_ENTRY_CODE,			"sendto_entry"},
			{FKT_SENDTO_EXIT_CODE,			"sendto_exit"},
			{FKT_RECVFROM_ENTRY_CODE,		"recvfrom_entry"},
			{FKT_RECVFROM_EXIT_CODE,		"recvfrom_exit"},
			{FKT_SHUTDOWN_ENTRY_CODE,		"shutdown_entry"},
			{FKT_SHUTDOWN_EXIT_CODE,		"shutdown_exit"},

			/* lamp routines */
			{FKT_LAMP_CREATE_ENTRY_CODE,	"lamp_create_entry"},
			{FKT_LAMP_CREATE_EXIT_CODE,		"lamp_create_exit"},
			{FKT_LAMP_BIND_ENTRY_CODE,		"lamp_bind_entry"},
			{FKT_LAMP_BIND_EXIT_CODE,		"lamp_bind_exit"},
			{FKT_LAMP_LISTEN_ENTRY_CODE,	"lamp_listen_entry"},
			{FKT_LAMP_LISTEN_EXIT_CODE,		"lamp_listen_exit"},
			{FKT_LAMP_ACCEPT_ENTRY_CODE,	"lamp_accept_entry"},
			{FKT_LAMP_ACCEPT_EXIT_CODE,		"lamp_accept_exit"},
			{FKT_LAMP_CONNECT_ENTRY_CODE,	"lamp_connect_entry"},
			{FKT_LAMP_CONNECT_EXIT_CODE,	"lamp_connect_exit"},
			{FKT_LAMP_SENDMSG_ENTRY_CODE,	"lamp_sendmsg_entry"},
			{FKT_LAMP_SENDMSG_EXIT_CODE,	"lamp_sendmsg_exit"},
			{FKT_LAMP_RECVMSG_ENTRY_CODE,	"lamp_recvmsg_entry"},
			{FKT_LAMP_RECVMSG_EXIT_CODE,	"lamp_recvmsg_exit"},
			{FKT_LAMP_DATA_RCV_ENTRY_CODE,	"lamp_data_rcv_entry"},
			{FKT_LAMP_DATA_RCV_EXIT_CODE,	"lamp_data_rcv_exit"},
			{FKT_LAMP_SIGNAL_RCV_ENTRY_CODE,"lamp_signal_rcv_entry"},
			{FKT_LAMP_SIGNAL_RCV_EXIT_CODE,	"lamp_signal_rcv_exit"},
			{FKT_LAMP_SHUTDOWN_ENTRY_CODE,	"lamp_shutdown_entry"},
			{FKT_LAMP_SHUTDOWN_EXIT_CODE,	"lamp_shutdown_exit"},
			{FKT_LAMP_RECVMSG_SCHEDI_CODE,	"lamp_recvmsg_sched_entry"},
			{FKT_LAMP_RECVMSG_SCHEDO_CODE,	"lamp_recvmsg_sched_exit"},

			/*	inet routines */
			{FKT_INET_CREATE_ENTRY_CODE,	"inet_create_entry"},
			{FKT_INET_CREATE_EXIT_CODE,		"inet_create_exit"},
			{FKT_INET_BIND_ENTRY_CODE,		"inet_bind_entry"},
			{FKT_INET_BIND_EXIT_CODE,		"inet_bind_exit"},
			{FKT_INET_LISTEN_ENTRY_CODE,	"inet_listen_entry"},
			{FKT_INET_LISTEN_EXIT_CODE,		"inet_listen_exit"},
			{FKT_INET_ACCEPT_ENTRY_CODE,	"inet_accept_entry"},
			{FKT_INET_ACCEPT_EXIT_CODE,		"inet_accept_exit"},
			{FKT_INET_STREAM_CONNECT_ENTRY_CODE,"inet_stream_connect_entry"},
			{FKT_INET_STREAM_CONNECT_EXIT_CODE,	"inet_stream_connect_exit"},
			{FKT_INET_SENDMSG_ENTRY_CODE,	"inet_sendmsg_entry"},
			{FKT_INET_SENDMSG_EXIT_CODE,	"inet_sendmsg_exit"},
			{FKT_INET_RECVMSG_ENTRY_CODE,	"inet_recvmsg_entry"},
			{FKT_INET_RECVMSG_EXIT_CODE,	"inet_recvmsg_exit"},
			{FKT_INET_SHUTDOWN_ENTRY_CODE,	"inet_shutdown_entry"},
			{FKT_INET_SHUTDOWN_EXIT_CODE,	"inet_shutdown_exit"},

			/*	net routines */
			{FKT_NET_RX_ACTION_ENTRY_CODE,	"net_rx_action_entry"},
			{FKT_NET_RX_ACTION_EXIT_CODE,	"net_rx_action_exit"},
//			{FKT_NET_BH_QUEUE_ENTRY_CODE, 	   "net_bh_queue_entry"},
//			{FKT_NET_BH_QUEUE_EXIT_CODE, 	   "net_bh_queue_exit"},
			{FKT_NETIF_RX_ENTRY_CODE,          "netif_rx_entry"},
            {FKT_NETIF_RX_EXIT_CODE,           "netif_rx_exit"},
			{FKT_NET_RX_DEQUEUE_ENTRY_CODE,    "net_rx_dequeue_entry"},
			{FKT_NET_RX_DEQUEUE_EXIT_CODE, 	   "net_rx_dequeue_exit"},

			/* sadhna end */

			/* miru start */

			/*	tcp routines */
			{FKT_TCP_SENDMSG_ENTRY_CODE,      "tcp_sendmsg_entry"},
			{FKT_TCP_SENDMSG_EXIT_CODE,       "tcp_sendmsg_exit"},
			{FKT_TCP_SEND_SKB_ENTRY_CODE,     "tcp_send_skb_entry"},
			{FKT_TCP_SEND_SKB_EXIT_CODE,      "tcp_send_skb_exit"},
			{FKT_TCP_TRANSMIT_SKB_ENTRY_CODE, "tcp_transmit_skb_entry"},
			{FKT_TCP_TRANSMIT_SKB_EXIT_CODE,  "tcp_transmit_skb_exit"},
			{FKT_TCP_V4_RCV_ENTRY_CODE,		"tcp_v4_rcv_entry"},
			{FKT_TCP_V4_RCV_EXIT_CODE,		"tcp_v4_rcv_exit"},
//			{FKT_TCP_RECVMSG_SCHEDI_CODE,	"tcp_data_wait_entry"},
//			{FKT_TCP_RECVMSG_SCHEDO_CODE,	"tcp_data_wait_exit"},
			{FKT_TCP_RCV_ESTABLISHED_ENTRY_CODE,"tcp_rcv_established_entry"},
			{FKT_TCP_RCV_ESTABLISHED_EXIT_CODE,"tcp_rcv_established_exit"},
			{FKT_TCP_DATA_ENTRY_CODE,          "tcp_data_entry"},
			{FKT_TCP_DATA_EXIT_CODE,           "tcp_data_exit"},
			{FKT_TCP_ACK_ENTRY_CODE,           "tcp_ack_entry"},
			{FKT_TCP_ACK_EXIT_CODE,            "tcp_ack_exit"},

			{FKT_TCP_RECVMSG_ENTRY_CODE,       "tcp_recvmsg_entry"},
			{FKT_TCP_RECVMSG_EXIT_CODE,        "tcp_recvmsg_exit"},
//			{FKT_TCP_V4_RCV_CSUM_ENTRY_CODE,   "tcp_v4_checksum_init_entry"},
//			{FKT_TCP_V4_RCV_CSUM_EXIT_CODE,    "tcp_v4_checksum_init_exit"},
			{FKT_TCP_RECVMSG_OK_ENTRY_CODE,    "tcp_recvmsg_ok_entry"},
			{FKT_TCP_RECVMSG_OK_EXIT_CODE,     "tcp_recvmsg_ok_exit"},
			{FKT_TCP_SEND_ACK_ENTRY_CODE,      "tcp_send_ack_entry"},
			{FKT_TCP_SEND_ACK_EXIT_CODE,       "tcp_send_ack_exit"},
//			{FKT_TCP_V4_RCV_HW_ENTRY_CODE,     "tcp_v4_rcv_hw_entry"},
//			{FKT_TCP_V4_RCV_HW_EXIT_CODE,      "tcp_v4_rcv_hw_exit"},
//			{FKT_TCP_V4_RCV_NONE_ENTRY_CODE,   "tcp_v4_rcv_none_entry"},
//			{FKT_TCP_V4_RCV_NONE_EXIT_CODE,    "tcp_v4_rcv_none_exit"},
			{FKT_WAIT_FOR_TCP_MEMORY_ENTRY_CODE,"wait_for_tcp_memory_entry"},
			{FKT_WAIT_FOR_TCP_MEMORY_EXIT_CODE,"wait_for_tcp_memory_exit"},
//			{FKT_TCP_RECVMSG_CSUM_ENTRY_CODE,  "tcp_recvmsg_csum_entry"},
//			{FKT_TCP_RECVMSG_CSUM_EXIT_CODE,   "tcp_recvmsg_csum_exit"},
//			{FKT_TCP_RECVMSG_CSUM2_ENTRY_CODE, "tcp_recvmsg_csum2_entry"},
//			{FKT_TCP_RECVMSG_CSUM2_EXIT_CODE,  "tcp_recvmsg_csum2_exit"},
			{FKT_TCP_SYNC_MSS_ENTRY_CODE,	   "tcp_sync_mss_entry"},
			{FKT_TCP_SYNC_MSS_EXIT_CODE,	   "tcp_sync_mss_exit"},

			/* udp routines */
			{FKT_UDP_SENDMSG_ENTRY_CODE,       "udp_sendmsg_entry"},
			{FKT_UDP_SENDMSG_EXIT_CODE,        "udp_sendmsg_exit"},
			{FKT_UDP_QUEUE_RCV_SKB_ENTRY_CODE, "udp_queue_rcv_skb_entry"},
			{FKT_UDP_QUEUE_RCV_SKB_EXIT_CODE,  "udp_queue_rcv_skb_exit"},
			{FKT_UDP_GETFRAG_ENTRY_CODE,       "udp_getfrag_entry"},
			{FKT_UDP_GETFRAG_EXIT_CODE,        "udp_getfrag_exit"},
			{FKT_UDP_RECVMSG_ENTRY_CODE,       "udp_recvmsg_entry"},
			{FKT_UDP_RECVMSG_EXIT_CODE,        "udp_recvmsg_exit"},
			{FKT_UDP_RCV_ENTRY_CODE,           "udp_rcv_entry"},
			{FKT_UDP_RCV_EXIT_CODE,            "udp_rcv_exit"},

			/*	ip routines */
			{FKT_IP_RCV_ENTRY_CODE,			"ip_rcv_entry"},
			{FKT_IP_RCV_EXIT_CODE,			"ip_rcv_exit"},
			{FKT_IP_QUEUE_XMIT_ENTRY_CODE,    "ip_queue_xmit_entry"},
			{FKT_IP_QUEUE_XMIT_EXIT_CODE,     "ip_queue_xmit_exit"},
			{FKT_IP_BUILD_XMIT_ENTRY_CODE,     "ip_build_xmit_entry"},
			{FKT_IP_BUILD_XMIT_EXIT_CODE,      "ip_build_xmit_exit"},
			{FKT_IP_LOCAL_DELIVER_ENTRY_CODE,  "ip_local_deliver_entry"},
			{FKT_IP_LOCAL_DELIVER_EXIT_CODE,   "ip_local_deliver_exit"},
			{FKT_IP_DEFRAG_ENTRY_CODE,         "ip_defrag_entry"},
			{FKT_IP_DEFRAG_EXIT_CODE,          "ip_defrag_exit"},
			{FKT_IP_FRAG_REASM_ENTRY_CODE,     "ip_frag_reasm_entry"},
			{FKT_IP_FRAG_REASM_EXIT_CODE,      "ip_frag_reasm_exit"},

			/*	ethernet cards */
			{FKT_TULIP_START_XMIT_ENTRY_CODE,    "tulip_start_xmit_entry"},
			{FKT_TULIP_START_XMIT_EXIT_CODE,     "tulip_start_xmit_exit"},
			{FKT_TULIP_INTERRUPT_ENTRY_CODE,     "tulip_interrupt_entry"},
			{FKT_TULIP_INTERRUPT_EXIT_CODE,      "tulip_interrupt_exit"},
			{FKT_TULIP_RX_ENTRY_CODE,            "tulip_rx_entry"},
			{FKT_TULIP_RX_EXIT_CODE,             "tulip_rx_exit"},
			{FKT_TULIP_TX_ENTRY_CODE,            "tulip_tx_entry"},
			{FKT_TULIP_TX_EXIT_CODE,             "tulip_tx_exit"},

			{FKT_VORTEX_INTERRUPT_ENTRY_CODE,    "vortex_interrupt_entry"},
			{FKT_VORTEX_INTERRUPT_EXIT_CODE,     "vortex_interrupt_exit"},
			{FKT_BOOMERANG_INTERRUPT_ENTRY_CODE, "boomerang_interrupt_entry"},
			{FKT_BOOMERANG_INTERRUPT_EXIT_CODE,  "boomerang_interrupt_exit"},
			{FKT_BOOMERANG_START_XMIT_ENTRY_CODE,"boomerang_start_xmit_entry"},
			{FKT_BOOMERANG_START_XMIT_EXIT_CODE, "boomerang_start_xmit_exit"},
			{FKT_BOOMERANG_RX_ENTRY_CODE,        "boomerang_rx_entry"},
			{FKT_BOOMERANG_RX_EXIT_CODE,         "boomerang_rx_exit"},
			{FKT_BOOMERANG_TX_ENTRY_CODE,        "boomerang_tx_entry"},
			{FKT_BOOMERANG_TX_EXIT_CODE,         "boomerang_tx_exit"},

			{FKT_DO_SOFTIRQ_ENTRY_CODE,          "do_softirq_entry"},
			{FKT_DO_SOFTIRQ_EXIT_CODE,           "do_softirq_exit"},

			{FKT_ACE_INTERRUPT_ENTRY_CODE,  "ace_interrupt_entry"},
			{FKT_ACE_INTERRUPT_EXIT_CODE,   "ace_interrupt_exit"},
			{FKT_ACE_RX_INT_ENTRY_CODE,     "ace_rx_int_entry"},
			{FKT_ACE_RX_INT_EXIT_CODE,      "ace_rx_int_exit"},
			{FKT_ACE_TX_INT_ENTRY_CODE,     "ace_tx_int_entry"},
			{FKT_ACE_TX_INT_EXIT_CODE,      "ace_tx_int_exit"},
			{FKT_ACE_START_XMIT_ENTRY_CODE, "ace_start_xmit_entry"},
			{FKT_ACE_START_XMIT_EXIT_CODE,  "ace_start_xmit_exit"},

			/* miru end */


			/*	the sys_xxx functions */
			/* achadda start */
			{FKT_READ_ENTRY_CODE,			"sys_read_entry"},
			{FKT_READ_EXIT_CODE,			"sys_read_exit"},

			/* vs start */
			{FKT_WRITE_ENTRY_CODE,			"sys_write_entry"},
			{FKT_WRITE_EXIT_CODE,			"sys_write_exit"},

			/* rdr */
			{FKT_LSEEK_ENTRY_CODE,			"sys_lseek_entry"},
			{FKT_LSEEK_EXIT_CODE,			"sys_lseek_exit"},

			/* vs start */
			{FKT_FSYNC_ENTRY_CODE,			"sys_fsync_entry"},
			{FKT_FSYNC_EXIT_CODE,			"sys_fsync_exit"},
			{FKT_NANOSLEEP_ENTRY_CODE,		"sys_nanosleep_entry"},
			{FKT_NANOSLEEP_EXIT_CODE,		"sys_nanosleep_exit"},


			/* rdr start */
//			{FKT_DO_SD_REQUEST_ENTRY_CODE,	"do_sd_request_entry"},
//			{FKT_DO_SD_REQUEST_EXIT_CODE,	"do_sd_request_exit"},
//			{FKT_REQUEUE_SD_REQUEST_ENTRY_CODE, "requeue_sd_request_entry"},
//			{FKT_REQUEUE_SD_REQUEST_EXIT_CODE, "requeue_sd_request_exit"},
			{FKT_SD_OPEN_ENTRY_CODE,		"sd_open_entry"},
			{FKT_SD_OPEN_EXIT_CODE,			"sd_open_exit"},
			{FKT_SD_RELEASE_ENTRY_CODE,		"sd_release_entry"},
			{FKT_SD_RELEASE_EXIT_CODE,		"sd_release_exit"},
			{FKT_SD_IOCTL_ENTRY_CODE,		"sd_ioctl_entry"},
			{FKT_SD_IOCTL_EXIT_CODE,		"sd_ioctl_exit"},
			{FKT_SCSI_DO_CMD_ENTRY_CODE,	"scsi_do_cmd_entry"},
			{FKT_SCSI_DO_CMD_EXIT_CODE,		"scsi_do_cmd_exit"},
			{FKT_SCSI_IOCTL_ENTRY_CODE,		"scsi_ioctl_entry"},
			{FKT_SCSI_IOCTL_EXIT_CODE,		"scsi_ioctl_exit"},
			{FKT_SCSI_DONE_ENTRY_CODE,		"scsi_done_entry"},
			{FKT_SCSI_DONE_EXIT_CODE,		"scsi_done_exit"},
			{FKT_SCSI_DISPATCH_CMD_ENTRY_CODE,	"scsi_dispatch_cmd_entry"},
			{FKT_SCSI_DISPATCH_CMD_EXIT_CODE,	"scsi_dispatch_cmd_exit"},
			{FKT_SCSI_RETRY_COMMAND_ENTRY_CODE,	"scsi_retry_command_entry"},
			{FKT_SCSI_RETRY_COMMAND_EXIT_CODE,	"scsi_retry_command_exit"},
			{FKT_SG_OPEN_ENTRY_CODE,		"sg_open_entry"},
			{FKT_SG_OPEN_EXIT_CODE,			"sg_open_exit"},
			{FKT_SG_RELEASE_ENTRY_CODE,		"sg_release_entry"},
			{FKT_SG_RELEASE_EXIT_CODE,		"sg_release_exit"},
			{FKT_SG_IOCTL_ENTRY_CODE,		"sg_ioctl_entry"},
			{FKT_SG_IOCTL_EXIT_CODE,		"sg_ioctl_exit"},
			{FKT_SG_READ_ENTRY_CODE,		"sg_read_entry"},
			{FKT_SG_READ_EXIT_CODE,			"sg_read_exit"},
			{FKT_SG_NEW_READ_ENTRY_CODE,	"sg_new_read_entry"},
			{FKT_SG_NEW_READ_EXIT_CODE,		"sg_new_read_exit"},
			{FKT_SG_WRITE_ENTRY_CODE,		"sg_write_entry"},
			{FKT_SG_WRITE_EXIT_CODE,		"sg_write_exit"},
			{FKT_SG_NEW_WRITE_ENTRY_CODE,	"sg_new_write_entry"},
			{FKT_SG_NEW_WRITE_EXIT_CODE,	"sg_new_write_exit"},
			{FKT_SG_GET_RQ_MARK_ENTRY_CODE,	"sg_get_rq_mark_entry"},
			{FKT_SG_GET_RQ_MARK_EXIT_CODE,	"sg_get_rq_mark_exit"},
			{FKT_SG_CMD_DONE_BH_ENTRY_CODE,	"sg_cmd_done_bh_entry"},
			{FKT_SG_CMD_DONE_BH_EXIT_CODE,	"sg_cmd_done_bh_exit"},
			{FKT_SCSI_SEP_IOCTL_ENTRY_CODE,	"scsi_sep_ioctl_entry"},
			{FKT_SCSI_SEP_IOCTL_EXIT_CODE,	"scsi_sep_ioctl_exit"},
			{FKT_SCSI_SEP_QUEUECOMMAND_ENTRY_CODE,"scsi_sep_queuecommand_entry"},
			{FKT_SCSI_SEP_QUEUECOMMAND_EXIT_CODE,"scsi_sep_queuecommand_exit"},
			/* rdr end */

			/* vs start */
//			{FKT_SEAGATE_INTERNAL_CMD_ENTRY_CODE, "seagate_internal_cmd_entry"},
//			{FKT_SEAGATE_INTERNAL_CMD_EXIT_CODE, "seagate_internal_cmd_exit"},
//			{FKT_QLOGIC_QCMND_ENTRY_CODE,	"qlogic_qcmnd_entry"},
//			{FKT_QLOGIC_QCMND_EXIT_CODE,	"qlogic_qcmnd_exit"},
//			{FKT_ISP_RET_STATUS_ENTRY_CODE,	"qlogic_ret_status_entry"},
//			{FKT_ISP_RET_STATUS_EXIT_CODE,	"qlogic_ret_status_exit"},
//			{AM53C974_QCMD_ENTRY_CODE,		"am53c974_qcmd_entry"},
//			{AM53C974_QCMD_EXIT_CODE,		"am53c974_qcmd_exit"},

			/*****  vs: old codes, need to be reassigned if used again *****
			{0x581, "ext2_sync_file_entry"},
			{0x681, "ext2_sync_file_exit"},
			{0x582, "sync_tind_entry"},
			{0x682, "sync_tind_exit"},
			{0x583, "sync_dind_entry"},
			{0x683, "sync_dind_exit"},
			{0x584, "sync_indirect_entry"},
			{0x684, "sync_indirect_exit"},
			{0x585, "sync_direct_entry"},
			{0x685, "sync_direct_exit"},
			{0x586, "sync_iblock_entry"},
			{0x686, "sync_iblock_exit"},
			{0x587, "sync_block_entry"},
			{0x687, "sync_block_exit"},
			*****/

//			{REQUE_DEVM_DEV_BLK_CODE,		"reque_prior_devm_dev_blk"},
//			{REQUE_BLK_CODE,				"reque_after_blk"},
			/* vs end */

			{0, NULL}
			};

#define	NTRAPS	20

char	*traps[NTRAPS+1] =	{
					"divide_error",						/*	 0 */
					"debug",							/*	 1 */
					"nmi",								/*	 2 */
					"int3",								/*	 3 */
					"overflow",							/*	 4 */
					"bounds",							/*	 5 */
					"invalid_op",						/*	 6 */
					"device_not_available",				/*	 7 */
					"double_fault",						/*	 8 */
					"coprocessor_segment_overrun",		/*	 9 */
					"invalid_TSS",						/*	10 */
					"segment_not_present",				/*	11 */
					"stack_segment",					/*	12 */
					"general_protection",				/*	13 */
					"page_fault",						/*	14 */
					"spurious_interrupt_bug",			/*	15 */
					"coprocessor_error",				/*	16 */
					"alignment_check",					/*	17 */
					"reserved",							/*	18 */
					"simd_coprocessor_error",			/*	19 */
					"machine_check"						/*	20 */
					};


#define NSYS_CALLS		232

char	*sys_calls[NSYS_CALLS+1] = {
			"sys_setup",					/* 0 */
			"sys_exit",	
			"sys_fork",	
			"sys_read",	
			"sys_write",	
			"sys_open",						/* 5 */
			"sys_close",	
			"sys_waitpid",	
			"sys_creat",	
			"sys_link",	
			"sys_unlink",					/* 10 */
			"sys_execve",	
			"sys_chdir",	
			"sys_time",	
			"sys_mknod",	
			"sys_chmod",					/* 15 */
			"sys_chown",	
			"sys_break",	
			"sys_stat",	
			"sys_lseek",	
			"sys_getpid",					/* 20 */
			"sys_mount",	
			"sys_umount",	
			"sys_setuid",	
			"sys_getuid",	
			"sys_stime",					/* 25 */
			"sys_ptrace",	
			"sys_alarm",	
			"sys_fstat",	
			"sys_pause",	
			"sys_utime",					/* 30 */
			"sys_stty",	
			"sys_gtty",	
			"sys_access",	
			"sys_nice",	
			"sys_ftime",					/* 35 */
			"sys_sync",	
			"sys_kill",	
			"sys_rename",	
			"sys_mkdir",	
			"sys_rmdir",					/* 40 */
			"sys_dup",	
			"sys_pipe",	
			"sys_times",	
			"sys_prof",	
			"sys_brk",						/* 45 */
			"sys_setgid",	
			"sys_getgid",	
			"sys_signal",	
			"sys_geteuid",	
			"sys_getegid",					/* 50 */
			"sys_acct",	
			"sys_phys",	
			"sys_lock",	
			"sys_ioctl",	
			"sys_fcntl",					/* 55 */
			"sys_mpx",	
			"sys_setpgid",	
			"sys_ulimit",	
			"sys_olduname",	
			"sys_umask",					/* 60 */
			"sys_chroot",	
			"sys_ustat",	
			"sys_dup2",	
			"sys_getppid",	
			"sys_getpgrp",					/* 65 */
			"sys_setsid",	
			"sys_sigaction",	
			"sys_sgetmask",	
			"sys_ssetmask",	
			"sys_setreuid",					/* 70 */
			"sys_setregid",	
			"sys_sigsuspend",	
			"sys_sigpending",	
			"sys_sethostname",	
			"sys_setrlimit",				/* 75 */
			"sys_getrlimit",	
			"sys_getrusage",	
			"sys_gettimeofday",	
			"sys_settimeofday",	
			"sys_getgroups",				/* 80 */
			"sys_setgroups",	
			"old_select",	
			"sys_symlink",	
			"sys_lstat",	
			"sys_readlink",					/* 85 */
			"sys_uselib",	
			"sys_swapon",	
			"sys_reboot",	
			"old_readdir",	
			"old_mmap",						/* 90 */
			"sys_munmap",	
			"sys_truncate",	
			"sys_ftruncate",	
			"sys_fchmod",	
			"sys_fchown",					/* 95 */
			"sys_getpriority",	
			"sys_setpriority",	
			"sys_profil",	
			"sys_statfs",	
			"sys_fstatfs",					/* 100 */
			"sys_ioperm",	
			"sys_socketcall",	
			"sys_syslog",	
			"sys_setitimer",	
			"sys_getitimer",				/* 105 */
			"sys_newstat",	
			"sys_newlstat",	
			"sys_newfstat",	
			"sys_uname",	
			"sys_iopl",						/* 110 */
			"sys_vhangup",	
			"sys_idle",	
			"sys_vm86",	
			"sys_wait4",	
			"sys_swapoff",					/* 115 */
			"sys_sysinfo",	
			"sys_ipc",	
			"sys_fsync",	
			"sys_sigreturn",	
			"sys_clone",					/* 120 */
			"sys_setdomainname",	
			"sys_newuname",	
			"sys_modify_ldt",	
			"sys_adjtimex",	
			"sys_mprotect",					/* 125 */
			"sys_sigprocmask",	
			"sys_create_module",	
			"sys_init_module",	
			"sys_delete_module",	
			"sys_get_kernel_syms",			/* 130 */
			"sys_quotactl",	
			"sys_getpgid",	
			"sys_fchdir",	
			"sys_bdflush",	
			"sys_sysfs",					/* 135 */
			"sys_personality",	
			"afs_syscall",
			"sys_setfsuid",	
			"sys_setfsgid",	
			"sys_llseek",					/* 140 */
			"sys_getdents",	
			"sys_select",	
			"sys_flock",	
			"sys_msync",	
			"sys_readv",					/* 145 */
			"sys_writev",	
			"sys_getsid",	
			"sys_fdatasync",	
			"sys_sysctl",	
			"sys_mlock",					/* 150 */
			"sys_munlock",	
			"sys_mlockall",	
			"sys_munlockall",	
			"sys_sched_setparam",	
			"sys_sched_getparam",			/* 155 */
			"sys_sched_setscheduler",	
			"sys_sched_getscheduler",	
			"sys_sched_yield",	
			"sys_sched_get_priority_max",	
			"sys_sched_get_priority_min",	/* 160 */
			"sys_sched_rr_get_interval",	
			"sys_nanosleep",	
			"sys_mremap",	
			"sys_setresuid",
			"sys_getresuid",				/* 165 */
			"sys_vm86",
			"sys_query_module",
			"sys_poll",
			"sys_nfsservctl",
			"sys_setresgid",				/* 170 */
			"sys_getresgid",
			"sys_prctl",
			"sys_rt_sigreturn",
			"sys_rt_sigaction",
			"sys_rt_sigprocmask",			/* 175 */
			"sys_rt_sigpending",
			"sys_rt_sigtimedwait",
			"sys_rt_sigqueueinfo",
			"sys_rt_sigsuspend",
			"sys_pread",					/* 180 */
			"sys_pwrite",
			"sys_chown",
			"sys_getcwd",
			"sys_capget",
			"sys_capset",					/* 185 */
			"sys_sigaltstack",
			"sys_sendfile",
			"sys_ni_syscall",				/* streams1 */
			"sys_ni_syscall",				/* streams2 */
			"sys_vfork",		            /* 190 */
			"sys_getrlimit",
			"sys_mmap2",
			"sys_truncate64",
			"sys_ftruncate64",
			"sys_stat64",					/* 195 */
			"sys_lstat64",
			"sys_fstat64",
			"sys_lchown",
			"sys_getuid",
			"sys_getgid",					/* 200 */
			"sys_geteuid",
			"sys_getegid",
			"sys_setreuid",
			"sys_setregid",
			"sys_getgroups",				/* 205 */
			"sys_setgroups",
			"sys_fchown",
			"sys_setresuid",
			"sys_getresuid",
			"sys_setresgid",				/* 210 */
			"sys_getresgid",
			"sys_chown",
			"sys_setuid",
			"sys_setgid",
			"sys_setfsuid",					/* 215 */
			"sys_setfsgid",
			"sys_pivot_root",
			"sys_mincore",
			"sys_madvise",
			"sys_getdents64",				/* 220 */
			"sys_fcntl64",
			"sys_ni_syscall",				/* reserved for TUX */
			"sys_ni_syscall",									   /* rdr */
			"sys_ni_syscall",									   /* rdr */
			"sys_fkt_setup",				/* 225 */			   /* rdr */
			"sys_fkt_endup",									   /* rdr */
			"sys_fkt_keychange",								   /* rdr */
			"sys_fkt_reset",									   /* rdr */
			"sys_fkt_getcopy",									   /* rdr */
			"sys_fkt_probe0",				/* 230 */			   /* rdr */
			"sys_fkt_probe1",									   /* rdr */
			"sys_fkt_probe2"				/* 232 */			   /* rdr */
			};


/*	user area to hold copy of system trace buffer */

/*	basetime[thiscpu] is first absolute time ever seen for thiscpu */
/*	lasttime[thiscpu] is last  absolute time ever seen for thiscpu */
/*	lastreltime[thiscpu] is last relative time ever seen for thiscpu */
/*	should have lastreltime = lasttime - basetime */
static u_64			*basetime, *basehigh, *lasttime;
static u_64			*lastreltime;

/*	start_time_cpu[thiscpu] is 1st relative time for thiscpu once taking stats*/
/*	end_time_cpu[thiscpu] is 1st relative time for thiscpu after taking stats */
static u_64			*start_time_cpu, *end_time_cpu;

/*	already_saw[thiscpu] is 0 before any slot for thiscpu is seen, else 1 */
/*	fun_time_cpu[thiscpu] is 1 after thiscpu has started taking stats,
							-1 after thiscpu has finished taking stats */
static int			*already_saw, *fun_time_cpu;
static unsigned int	lastcpu = -1;
static unsigned int	*lastpid;
static clock_t		start_jiffies = 0, stop_jiffies = 0;
static time_t		start_time = 0, stop_time = 0;
static int			fd;
static unsigned long pid = 0;
static unsigned long kidpid = 0;

/* set to 1 by -p switch or as soon as a task with kidpid is found */
/* when 1, allows printing of probe items */
static int kidflag = 0;

/* set to 1 by -a switch to take statistics for all probe items */
static int all_stats = 0;

/* fun_time is initially 0.
 * set to 1 when hit first sys_call entry to times for kidpid task
 *			at which time start_time_glob is set to reltime.
 *			or if all_stats is set, on the first probe of any kind.
 * while 1, allows accumulation of entry/exit statistics
 * set to -1 when hit sys_call entry to times for kidpid task and
 *			it is already set to 1.  end_time_glob is set to reltime then.
 *			of if all_stats is set, on first fkt_keychange.
 */
static int fun_time = 0;

static struct code_list_item	*irq_list = NULL;
static unsigned int nirqs = 0;
static unsigned int ncpus = 0;



void load_irqs( void )
	{
	struct code_list_item	*ptr;
	unsigned int			i, k, n;


	printf("\n");
	for( n = 0;  n < nirqs;  n++ )
		{/* build circular list with 1 item for each interrupt */
		ptr = (struct code_list_item *)malloc(sizeof(struct code_list_item));
		read(fd, (void *)&ptr->code, sizeof(ptr->code));
		read(fd, (void *)&i, sizeof(i));
		k = (i + 3) & ~3;
		ptr->name = malloc(k);
		read(fd, (void *)ptr->name, k);
		ptr->name[i-1] = '\0';
		if( irq_list == NULL )
			ptr->link = ptr;
		else
			{
			ptr->link = irq_list->link;
			irq_list->link = ptr;
			}
		irq_list = ptr;
		printf("irq %2u: %s\n", ptr->code, ptr->name);
		}
	printf("\n");
	if( (ptr = irq_list) != NULL )
		{/* make circular list into singly-linked, NULL-terminated list */
		irq_list = irq_list->link;
		ptr->link = NULL;
		}
	}

char *find_irq( unsigned int code )
	{
	struct code_list_item	*ptr;

	for( ptr = irq_list;  ptr != NULL;  ptr = ptr->link )
		{
		if( ptr->code == code )
			return ptr->name;
		}
	return "unknown irq";
	}

char *find_name( unsigned int code, int keep_entry )
	{
	struct code_name	*ptr;
	static char			local_buf[128];
	int					len, elen;

	for( ptr = code_table;  ptr->code != 0;  ptr++ )
		if( ptr->code == code )
			{
			if( keep_entry )
				return ptr->name;
			else
				{/* caller wants _entry stripped off end of name */
				len = strlen(ptr->name);
				elen = strlen("_entry");
				if( len > elen  && strcmp(&ptr->name[len-elen], "_entry") == 0 )
					{
					strcpy(local_buf, ptr->name);
					local_buf[len-elen] = '\0';
					return local_buf;
					}
				else
					return ptr->name;
				}
			}
	return "unknown code";
	}

int find_syscall( int code, char **category, char **name, int *should_be_hex )
	{
	int		z;

	*should_be_hex = 0;
	if( code < FKT_SYS_CALL_LIMIT_CODE )
		{/* this is a true system call */
		z = code;
		*should_be_hex = 1;
		*category = "sys_call";
		*name = sys_calls[z];
		}
	else if( code < FKT_TRAP_LIMIT_CODE )
		{/* this is a trap */
		z = code-FKT_SYS_CALL_LIMIT_CODE;
		if( z > NTRAPS )
			z = NTRAPS;
		*category = "trap";
		*name = traps[z];
		}
	else
		{
		z = code-FKT_TRAP_LIMIT_CODE;
		*category = "IRQ";
		*name = find_irq(z);
		}
	return z;
	}


static struct pid_cycles_item	*pid_cycles_list = NULL;

void update_pid_cycles( u_64 cycles, unsigned int pid )
	{
	struct pid_cycles_item	*ptr;

	for( ptr = pid_cycles_list;  ptr != NULL;  ptr = ptr->link )
		{
		if( ptr->pid == pid )
			{/* found item for this pid, just increment its cycles */
			ptr->total_cycles += cycles;
			fprintf(stdbug, "add %llu cycles to pid %u, total now %llu\n",
											cycles, pid, ptr->total_cycles);
			return;
			}
		}
	/* loop finishes when item for this pid not found, make a new one */
	if( (ptr = (struct pid_cycles_item *)malloc(sizeof(struct pid_cycles_item)))
					== NULL )
		{
		fprintf(stderr, "unable to malloc pid_cycles_item\n");
		exit(EXIT_FAILURE);
		}
	ptr->pid = pid;
	ptr->total_cycles = cycles;
	ptr->link = pid_cycles_list;		/* add to front of list */
	pid_cycles_list = ptr;
	fprintf(stdbug,"create new cycles for pid %u, initial %llu\n",pid,cycles);
	}


void print_cycles( u_64 ratio )
	{
	struct pid_cycles_item	*ptr;
	double					t, r, total, jtot, ptot;

	r = (double)ratio;
	total = 0.0;
	jtot = 0.0;
	ptot = 0.0;
	printf("%6s %12s %10s %11s\n", "pid", "cycles", "jiffies", "percent");
	printf("\n");
	for( ptr = pid_cycles_list;  ptr != NULL;  ptr = ptr->link )
		{
		t = ((double)ptr->total_cycles);
		total += t;
		jtot += t/r;
		}
	for( ptr = pid_cycles_list;  ptr != NULL;  ptr = ptr->link )
		{
		t = ((double)ptr->total_cycles);
		printf("%6u %12.0f %10.2f %10.2f%%\n", ptr->pid, t, t/r, t*100.0/total);
		ptot += t*100.0/total;
		}
	printf("%6s %12.0f %10.2f %10.2f%%\n", "TOTAL", total, jtot, ptot);
	}


/* first probe for this pid, create and initialize a new stack item for it */
struct v3_stack_item *create_stack( unsigned int thispid, unsigned int thiscpu,
															u_64 start_time )
	{
	struct v3_stack_item	*v3_ptr;

	fprintf(stdbug, "create_stack for pid %u, cpu %u, rel time %llu\n",
												thispid, thiscpu, start_time);

	v3_ptr = (struct v3_stack_item *)malloc(sizeof(struct v3_stack_item));
	if( v3_ptr == NULL )
		{
		fprintf(stderr, "=== can't malloc stack for pid %d\n", thispid);
		fprintf(stdout, "=== can't malloc stack for pid %d\n", thispid);
		exit(EXIT_FAILURE);
		}
	v3_ptr->v3_pid = thispid;
	v3_ptr->v3_cpu = thiscpu;
	v3_ptr->v3_process_cycles = 0LL;
	v3_ptr->v3_pos = 0;
	v3_ptr->v3_code[0] = 0;
	if( v3_stack_first == NULL )
		{/* this is the first stack ever created */
		v3_stack_first = v3_ptr;
		}
	if( v3_stack_ptr[thiscpu] == NULL )
		{/* this is the first stack for this cpu */
		v3_ptr->v3_start_cycle[0] = start_time;
		}
	else
		{/* initialize new stack with time of item at top of current stack */
		v3_ptr->v3_start_cycle[0] =
		v3_stack_ptr[thiscpu]->v3_start_cycle[v3_stack_ptr[thiscpu]->v3_pos];
		}

	v3_ptr->v_index_arr_pos = 0;
	v3_ptr->v_index_array[0] = 0;

	v3_ptr->v2_pos = 0;
	v3_ptr->v2_start_cycle[0] = 0LL;
	v3_ptr->v2_code[0] = 0;

	v3_ptr->v3_next = v3_stack_head;	/* push on stack list */
	v3_stack_head = v3_ptr;
	return v3_ptr;
	}


void update_stack( struct v3_stack_item	*v3_ptr, u_64 delta )
	{
	unsigned int	i;

	/* move up start time of all previous in nesting */
	for( i = 0;  i < v3_ptr->v3_pos;  i++ )
		v3_ptr->v3_start_cycle[i] += delta;

	/* then pop the stack */
	v3_ptr->v3_pos--;
	fprintf(stdbug, "pop stack for pid %u, pos %d\n",
				v3_ptr->v3_pid, v3_ptr->v3_pos);
	}


/*	here to flush out the v3 stack and correct entries in v1 stack */
/*	called only when ret_from_sys_call finds non-empty stack */
void fixup_stack( void )
	{
	int		i, k;

	while( v3_stack_ptr[lastcpu]->v3_pos > 0 )
		{
		/*	i gets code at top of the nesting stack for this pid */
		i = v3_stack_ptr[lastcpu]->v3_code[v3_stack_ptr[lastcpu]->v3_pos];

    	if( i > FKT_UNSHIFTED_LIMIT_CODE )
			{/*	this is a function entry or exit */
			/*	search the v array to see if we have seen this code before */
			for( k = 0;  k < v_pos;  k++ )
				{
				if ( i == v[k].code )
					{/* this code matches the code in v array, null it out */
					printf("=== NULL out v[%d].code, was %04x: %s\n",
												k, i, find_name(i, 1));
					v[k].code = 0;
					if( v3_stack_ptr[lastcpu]->v_index_arr_pos > 0  &&
						v3_stack_ptr[lastcpu]->v_index_array
							[v3_stack_ptr[lastcpu]->v_index_arr_pos - 1] == k )
						{/* in luck, pop off v_index_array stack top */
						v3_stack_ptr[lastcpu]->v_index_arr_pos--;
						if( k == v_pos - 1 )
							{/* more luck, pop off v stack top */
							v_pos--;
							}
						}
					break;
					}
				}
			}

		/* pop the nesting stack for this pid */
		v3_stack_ptr[lastcpu]->v3_pos--;
		}
	}


void update_function( u_64 delta, int i )
	{
	int				j, k;


	j = -1;
	for( k = 0;  k < MAX_FUNS;  k++ )
		{
		if( i == v3_function_code[k] )
			{/* this is the match, update counters for this function */
			break;
			}
		else if( j < 0  &&  v3_function_code[k] == 0 )
			{/* remember this empty slot in case we need to use it */
			j = k;
			}
		}
	if( k >= MAX_FUNS )
		{/* first occurrence of this entry/exit pair */
		if( j < 0 )
			{
			fprintf(stderr,"=== no room for function entry code %04x\n",i);
			fprintf(stdout,"=== no room for function entry code %04x\n",i);
			return;
			}
		else
			{/* store this new code here */
			v3_function_code[j] = i;
			k = j;
			if( actual_max_v3_function < j )
				actual_max_v3_function = j;
			}
		}

	v3_function_cycles[k] += delta;
	v3_function_count[k] += 1;

	fprintf(stdbug, "update function %s by %llu, total now %llu\n",
								find_name(i,1), delta, v3_function_cycles[k]);
	}

/* setup stack for this pid */
void setup_stack( u_64 cyc_time, unsigned int thispid )
	{
    u_64	delta, off_cyc_time;


	fprintf(stdbug, "setup_stack for pid %u, cpu %u\n", thispid, lastcpu);

	/*	cyc_time is current (relative) time since the starting base time */
	off_cyc_time = cyc_time;


	if( v3_stack_ptr[lastcpu] == NULL )
		{/* first time through here, we need a stack */
		v3_stack_ptr[lastcpu] = create_stack(thispid, lastcpu, cyc_time);
		}
	else if( v3_stack_ptr[lastcpu]->v3_pid != thispid )
		{/* next probe not the current pid, find a stack for the new pid */
		struct v3_stack_item	*v3_ptr;

		if( v3_stack_ptr[lastcpu]->v3_code[v3_stack_ptr[lastcpu]->v3_pos] ==
															FKT_SWITCH_TO_CODE )
			{/* last instruction was a switch_to, so use its time */
			off_cyc_time = v3_stack_ptr[lastcpu]->v3_start_cycle
												[v3_stack_ptr[lastcpu]->v3_pos];
			}
		else
			{/* top of current stack is not switch_to, should be */
			fprintf(stderr,
						"=== top of current stack for pid %u not switch_to\n",
						v3_stack_ptr[lastcpu]->v3_pid);
			fprintf(stdout,
						"=== top of current stack for pid %u not switch_to\n",
						v3_stack_ptr[lastcpu]->v3_pid);
			}

		/*	search list of existing items to find one for this pid */
		for( v3_ptr = v3_stack_head;  v3_ptr != NULL; v3_ptr = v3_ptr->v3_next )
			{
			if( v3_ptr->v3_pid == thispid  &&
								(v3_ptr->v3_cpu == lastcpu || thispid != 0) )
				{/* found an existing stack for the new pid */
				fprintf(stdbug, "go back to stack for pid %u, cpu %u, pos %d\n",
									thispid, v3_ptr->v3_cpu, v3_ptr->v3_pos);
				if( v3_ptr->v3_code[v3_ptr->v3_pos] == FKT_SWITCH_TO_CODE )
					{/* top of previous stack is switch_to, pop it off */
					v3_stack_ptr[lastcpu] = v3_ptr;
					delta = off_cyc_time -
										v3_ptr->v3_start_cycle[v3_ptr->v3_pos];
					update_function(delta, FKT_SWITCH_TO_CODE);
					/* move up start of all previous in nesting and pop stack */
					update_stack(v3_ptr, delta);
					}
				else if( v3_ptr->v3_pos != 0 )
					{
					fprintf(stderr,
							"=== previous stack does not have switch_to on top,"
							" pos = %d\n", v3_ptr->v3_pos);
					fprintf(stdout,
							"=== previous stack does not have switch_to on top,"
							" pos = %d\n", v3_ptr->v3_pos);
					}
				else /* v3_ptr->v3_pos is 0, previous stack is empty */
					{/* reinit previous stack with time of item at top of
						current stack */
					v3_ptr->v3_start_cycle[0] = v3_stack_ptr[lastcpu]->
								v3_start_cycle[v3_stack_ptr[lastcpu]->v3_pos];
					}
				v3_ptr->v3_cpu = lastcpu;	/* remember last cpu to work here */
				break;
				}
			}
		if( v3_ptr == NULL )
			{/* first probe for this pid, create and initialize a new stack */
			v3_ptr = create_stack(thispid, lastcpu, cyc_time);
			}
		v3_stack_ptr[lastcpu] = v3_ptr;
		}
	}

/* exiting the corresponding entry in v array at slot k */
void exit_function( struct v3_stack_item *v3_ptr, int k,
										unsigned int thispid, u_64 cyc_time )
	{
	int		i, j;
	u_64	cycdiff;

	/* find top of cycles_array stack
	   (start time of last call to this function) */
	for ( i = 0; i < MAX_EACH_FUN_RECUR; i++ )
		{
		if ( v[k].cycles_array[i] == 0LL )
			break;
		}
	if( i <= 0 )
		{/* no open (unexited) call to this function !?! */
		fprintf(stderr, "=== no open call to function %s\n",
										find_name(v[k].code,1));
		fprintf(stdout, "=== no open call to function %s\n",
										find_name(v[k].code,1));
		}
	else if( v[k].caller_pid[i-1] != thispid )
		{/* last call was done by someone else */
		fprintf(stderr, "=== last call to function %s in pid %u, "
			"not pid %u\n",
			find_name(v[k].code,1), v[k].caller_pid[i-1], thispid);
		fprintf(stdout, "=== last call to function %s in pid %u, "
			"not pid %u\n",
			find_name(v[k].code,1), v[k].caller_pid[i-1], thispid);
		}
	else
		{/*	compute time spent in last call to this function */
		cycdiff = cyc_time - v[k].cycles_array[i - 1];

		/* add this time to total time ever spent in function */
		if( v[k].code == FKT_NANOSLEEP_ENTRY_CODE )
			fprintf(stdbug,
				"add %llu to nanosleep_entry v[%d].total %llu\n",
											cycdiff, k, v[k].total);
		if( v[k].code == FKT_NANOSLEEP_EXIT_CODE )
			fprintf(stdbug,
				"add %llu to nanosleep_exit v[%d].total %llu\n",
											cycdiff, k, v[k].total);
		v[k].total += cycdiff;
			
		if (k == v3_ptr->v_index_array[v3_ptr->v_index_arr_pos - 1])
			{
			if (v3_ptr->v_index_arr_pos >= 2 )
			{
				j = v3_ptr->v_index_array[v3_ptr->v_index_arr_pos - 2];
			   v[j].called_funs_cycles[k] += cycdiff;  
			}
									 
			v3_ptr->v_index_array[v3_ptr->v_index_arr_pos]=-1;
			v3_ptr->v_index_arr_pos--;
			}

		/* "pop" the stack of calls to this function */
		v[k].cycles_array[i - 1] = 0LL;
		} 
	}


/* code for return from kernel functions */
/* this is the exit corresponding to a previously stacked entry */
/* or a switch back to previously stacked switch from */
void return_from_function( struct v3_stack_item	*v3_ptr, int code,
																u_64 cyc_time )
	{
	u_64	delta;

	delta = cyc_time - v3_ptr->v3_start_cycle[v3_ptr->v3_pos];
	update_function(delta, code);
	/* move up start time of all previous in nesting and pop stack */
	update_stack(v3_ptr, delta);
	}

/* code for return from system calls */
void return_from_sys_call( struct v3_stack_item	*v3_ptr, int code,
																u_64 cyc_time )
	{
	char	*category, *name;
	int		should_be_hex;
	u_64	delta;

	if( v3_ptr->v2_pos > 0 )
		{/* a syscall, trap or IRQ was previously stacked, measure it */
		syscall_cycles[v3_ptr->v2_code[v3_ptr->v2_pos]] +=
							cyc_time - v3_ptr->v2_start_cycle[v3_ptr->v2_pos];
		syscall_count[v3_ptr->v2_code[v3_ptr->v2_pos]] += 1;
		v3_ptr->v2_pos--;
		}

	if( v3_ptr->v3_pos>0 && code<FKT_UNSHIFTED_LIMIT_CODE )
		{/* stacked code was a trap/syscall/IRQ */
		find_syscall(code, &category, &name, &should_be_hex);
		delta = cyc_time - v3_ptr->v3_start_cycle[v3_ptr->v3_pos];
		v3_syscall_cycles[code] +=  delta;
		v3_syscall_count[code] += 1;
		fprintf(stdbug, "update %s %s by %llu, total now %llu\n",
						category, name, delta, v3_syscall_cycles[code]);

		/* move up start time of all previous in nesting & pop stack */
		update_stack(v3_ptr, delta);
		}
	else if( v3_ptr->v3_pos == 0 )
		{/* return to a process that entered the kernel before we
			started to record trace info, just ignore this return */
		fprintf(stdbug, "unmatched return from sys_call, ignored\n");
		}
	else
		{
		fprintf(stderr, "=== return from system call improperly nested: "
										"code %04x, depth %d\n",
										code,v3_ptr->v3_pos);
		fprintf(stdout, "=== return from system call improperly nested: "
										"code %04x, depth %d\n",
										code,v3_ptr->v3_pos);
		fixup_stack();
		}
	}

/*	here to close a non-empty v3 stack (v3_pos>0) when we stop taking stats */
/*	cyc_time is current (relative) time since the starting base time */
void close_stack( struct v3_stack_item	*v3_ptr, unsigned int thispid,
																u_64 cyc_time )
	{
	int		k, code, z, should_be_hex;
	char	*category, *name;
    

	fprintf(stdbug,
	"close stack for pid %4u, v3_pos %u, cyc_time %llu, process cycles %llu\n",
		thispid, v3_ptr->v3_pos, cyc_time,
		v3_ptr->v3_process_cycles);

	if( v3_ptr->v3_code[v3_ptr->v3_pos] != FKT_SWITCH_TO_CODE )
		{/* top of current stack is not switch_to, should be */
		fprintf(stderr,
					"=== pid %4u final stack top not switch_to\n",
					v3_ptr->v3_pid);
		fprintf(stdout,
					"=== pid %4u final stack top not switch_to\n",
					v3_ptr->v3_pid);
		}
	else
		{/* top of current stack is switch_to, use its time and pop it */
		cyc_time = v3_ptr->v3_start_cycle[v3_ptr->v3_pos];
		v3_ptr->v3_pos--;
		}

	while( v3_ptr->v3_pos > 0 )
		{
		/*	code gets code at top of this stack */
		code = v3_ptr->v3_code[v3_ptr->v3_pos];
		fprintf(stdbug, "v3[%u] ", v3_ptr->v3_pos);
    	if( code < FKT_UNSHIFTED_LIMIT_CODE )
			{/* this is a trap/syscall/irq entry */
			z = find_syscall(code, &category, &name, &should_be_hex);
			fprintf(stdbug, "%8s ", category);
			if( should_be_hex )
				fprintf(stdbug, "%04x", z);
			else
				fprintf(stdbug, "%4d", z);
			fprintf(stdbug, " %04x %28s\n", code, name);
			}
		else
			fprintf(stdbug, "%13s %04x %28s\n",
										"function", code, find_name(code,1));
    	if( code == FKT_SWITCH_TO_CODE )
			{/* instruction was a switch_to, use its time and pop it off */
			cyc_time = v3_ptr->v3_start_cycle[v3_ptr->v3_pos];
			v3_ptr->v3_pos--;
			}
    	else if( code < FKT_UNSHIFTED_LIMIT_CODE )
			{/* this is trap/syscall/irq */
			return_from_sys_call(v3_ptr, code, cyc_time);
			}
		else if( (code & FKT_GENERIC_EXIT_OFFSET) == 0 )
			{/* this is an exit code, not an entry code */
			fprintf(stderr, "=== exit code on stack: "
							"code %04x, %s\n", code, find_name(code, 1));
			fprintf(stdout, "=== exit code on stack: "
							"code %04x, %s\n", code, find_name(code, 1));
			/* now ignore it */
			v3_ptr->v3_pos--;
			}
		else
			{/* this is an entry code, force the exit code */
			return_from_function(v3_ptr,code+FKT_GENERIC_EXIT_OFFSET,cyc_time);
    		for( k = 0;  k < v_pos;  k++ )
    			{
        		if( code == v[k].code )
        			{/* v[k] is the entry code that matches this entry code */
					exit_function(v3_ptr, k, thispid, cyc_time);
					break;
					}
				}
			}
		}
	v3_ptr->v3_process_cycles =
							cyc_time-v3_ptr->v3_start_cycle[0];

	fprintf(stdbug, "pid %4u new process cycles %llu\n\n",
							thispid, v3_ptr->v3_process_cycles);
	}

void my_print_fun( unsigned int code, u_64 cyc_time,
								unsigned int thispid, int param1, int param2 )
	{
    int		i, j, k;

    
	setup_stack(cyc_time, thispid);

	/*	now guaranteed that v3_stack_ptr[lastcpu] points to stack for this pid*/
	/*	i gets code at top of this stack */
	i = v3_stack_ptr[lastcpu]->v3_code[v3_stack_ptr[lastcpu]->v3_pos];
    if( code < FKT_UNSHIFTED_LIMIT_CODE )
    	{/* this is a trap/syscall/irq entry, put it in v2 stack. */
		if( v3_stack_ptr[lastcpu]->v2_pos >= MAX_UNSHIFTED_FUN_RECUR - 1 )
			{
			fprintf(stderr, "=== trap/syscall/irq stack overflow\n");
			fprintf(stdout, "=== trap/syscall/irq stack overflow\n");
			}
		else
			{
			v3_stack_ptr[lastcpu]->v2_pos++;
			if( actual_max_v2_pos < v3_stack_ptr[lastcpu]->v2_pos )
				actual_max_v2_pos = v3_stack_ptr[lastcpu]->v2_pos;
			v3_stack_ptr[lastcpu]->v2_start_cycle[v3_stack_ptr[lastcpu]->v2_pos]
																	= cyc_time;
			v3_stack_ptr[lastcpu]->v2_code[v3_stack_ptr[lastcpu]->v2_pos]=code;
			if( v3_stack_ptr[lastcpu]->v3_pos >= MAX_SHIFTED_FUN_RECUR - 1 )
				{
				fprintf(stderr, "=== nesting stack overflow\n");
				fprintf(stdout, "=== nesting stack overflow\n");
				}
			else
				{/* push new item onto v3_stack */
				v3_stack_ptr[lastcpu]->v3_pos++;
				if( actual_max_v3_pos < v3_stack_ptr[lastcpu]->v3_pos )
					actual_max_v3_pos = v3_stack_ptr[lastcpu]->v3_pos;
				v3_stack_ptr[lastcpu]->v3_start_cycle
									[v3_stack_ptr[lastcpu]->v3_pos] = cyc_time;
				v3_stack_ptr[lastcpu]->v3_code[v3_stack_ptr[lastcpu]->v3_pos] =
																		code;
				fprintf(stdbug,
							"push stack for pid %u, pos %d, code %04x, "
							"trap/syscall/irq\n",
							thispid, v3_stack_ptr[lastcpu]->v3_pos, code);
				}
			}
        return;
    	}
    else if( code == FKT_RET_FROM_SYS_CALL_CODE )
    	{/* code for return from system calls */
		return_from_sys_call(v3_stack_ptr[lastcpu], i, cyc_time);
		return;
    	}
	else
		{/* code for a function entry/exit, including switch_to */
    	if( ((code & FKT_GENERIC_EXIT_OFFSET) == 0  &&
							code == i + FKT_GENERIC_EXIT_OFFSET) ||
			(code == FKT_SWITCH_TO_CODE  &&  i == FKT_SWITCH_TO_CODE 
								&&  param1 == v3_stack_ptr[lastcpu]->v3_pid) )
			{/* this is the exit corresponding to a previously stacked entry */
			/* or a switch back to previously stacked switch from */
			return_from_function(v3_stack_ptr[lastcpu], i, cyc_time);
			}
		else
			{/* this should be an entry code or unmatched switch to*/
			if( (code & FKT_GENERIC_EXIT_OFFSET) == 0 )
				{/* this is an exit code, not an entry code */
				fprintf(stderr, "=== exiting from uncalled function: "
								"code %04x, %s\n", code, find_name(code, 1));
				fprintf(stdout, "=== exiting from uncalled function: "
								"code %04x, %s\n", code, find_name(code, 1));
				/* now ignore it */
				return;
				}
			if( v3_stack_ptr[lastcpu]->v3_pos >= MAX_SHIFTED_FUN_RECUR - 1 )
				{
				fprintf(stderr, "=== nesting stack overflow\n");
				fprintf(stdout, "=== nesting stack overflow\n");
				}
			else
				{
				v3_stack_ptr[lastcpu]->v3_pos++;
				if( actual_max_v3_pos < v3_stack_ptr[lastcpu]->v3_pos )
					actual_max_v3_pos = v3_stack_ptr[lastcpu]->v3_pos;
				v3_stack_ptr[lastcpu]->v3_start_cycle
									[v3_stack_ptr[lastcpu]->v3_pos] = cyc_time;
				v3_stack_ptr[lastcpu]->v3_code[v3_stack_ptr[lastcpu]->v3_pos] =
																		code;
				fprintf(stdbug,
							"push stack for pid %u, pos %d, code %04x, "
							"%s\n",
							thispid, v3_stack_ptr[lastcpu]->v3_pos, code,
							code==FKT_SWITCH_TO_CODE?"switch_to":"function");
				if( code == FKT_SWITCH_TO_CODE )
					{/* should set up stack for new pid here */
					setup_stack(cyc_time, param1);
					}
				}
			}
		}
    if( code == FKT_SWITCH_TO_CODE )
		{
		return;
		}

	/*	this is a function entry or matched exit */
	/*	search the v array to see if we have seen this code before */
    for( k = 0;  k < v_pos;  k++ )
    	{
        if( code == v[k].code  ||  code == v[k].code + FKT_GENERIC_EXIT_OFFSET )
        	{/* v[k] is the entry code that matches this code's entry or exit */
            if ( code == v[k].code )
            	{/* this code is the entry code */
                
				/*	push pointer to this code's v array slot on v_index_array */
                v3_stack_ptr[lastcpu]->v_index_array
									[v3_stack_ptr[lastcpu]->v_index_arr_pos]=k;
                if ( v3_stack_ptr[lastcpu]->v_index_arr_pos >= 1 )
					{
					j = v3_stack_ptr[lastcpu]->v_index_array
									[v3_stack_ptr[lastcpu]->v_index_arr_pos-1];
					v[j].called_funs_ctr[k]++;
					}
				v3_stack_ptr[lastcpu]->v_index_arr_pos++;
                
				/* find top of cycles_array stack in v slot for this function */
				for ( i = 0; i < MAX_EACH_FUN_RECUR; i ++ )
					{
					if ( v[k].cycles_array[i] == 0LL )
						break;
					}

				if( i >= MAX_EACH_FUN_RECUR )
					{
					fprintf(stderr, "=== too many open calls to function %s, "
									"max of %d\n",
									find_name(v[k].code,1), MAX_EACH_FUN_RECUR);
					fprintf(stdout, "=== too many open calls to function %s, "
									"max of %d\n",
									find_name(v[k].code,1), MAX_EACH_FUN_RECUR);
					exit(EXIT_FAILURE);
					}

				if( actual_max_cycles_pos < i )
					actual_max_cycles_pos = i;
				/* time of last call to this function at top of cycles_array */
				v[k].cycles_array[i] = cyc_time;
				v[k].caller_pid[i] = thispid;

				/* count total number of calls to this function */
				v[k].counter++;

				if( v[k].code == FKT_NANOSLEEP_ENTRY_CODE  ||
					v[k].code == FKT_NANOSLEEP_EXIT_CODE )
					{
					fprintf(stdbug, "v[%d]: %s, code=%04x, counter=%d, "
							"cycles_array[%d]=%llu, caller_pid[%d]=%d\n",
							k, find_name(v[k].code,1), v[k].code, v[k].counter,
							i, v[k].cycles_array[i], i, v[k].caller_pid[i]);
					}
				}                
			else
				{/* this code is the exit corresponding to entry k in v array */
				/* find top of cycles_array stack
				   (start time of last call to this function) */
				exit_function(v3_stack_ptr[lastcpu], k, thispid, cyc_time);
				}
            return;
        	}     
    	} 

	/* if loop terminates this is first time we have seen this code */
	/* initialize and push a new slot at top of v array */
	if( v_pos >= MAX_FUNS )
		{/* v stack overflow */
		fprintf(stderr, "=== too many unique functions called, max of %d\n",
																	MAX_FUNS);
		fprintf(stdout, "=== too many unique functions called, max of %d\n",
																	MAX_FUNS);
		exit(EXIT_FAILURE);
		}
	if( actual_max_v_pos < v_pos )
		actual_max_v_pos = v_pos;
	v[v_pos].code = code; 
	v[v_pos].total = 0;
	v[v_pos].counter = 1;
	for ( i = 0; i < MAX_EACH_FUN_RECUR; i ++)
		{
		v[v_pos].cycles_array[i] = 0LL;
		v[v_pos].caller_pid[i] = 0;
		}
	for(a = 0; a < MAX_FUNS; a++ )
		{
		v[v_pos].called_funs_cycles[a] = 0LL;
		v[v_pos].called_funs_ctr[a] = 0;
		}

	/* also push pointer to this called function on v_index_array */
	v3_stack_ptr[lastcpu]->v_index_array[v3_stack_ptr[lastcpu]->v_index_arr_pos]
																		= v_pos;
	if ( v3_stack_ptr[lastcpu]->v_index_arr_pos >= 1 )
		{
		j = v3_stack_ptr[lastcpu]->v_index_array
								[v3_stack_ptr[lastcpu]->v_index_arr_pos - 1];
		if (v[j].called_funs_ctr[k] == 0 )
			v[j].called_funs_ctr[k]++;
		}
	v3_stack_ptr[lastcpu]->v_index_arr_pos++;

	/* save pid of caller and time of this call to this function */
	v[v_pos].cycles_array[0] = cyc_time;
	v[v_pos].caller_pid[0] = thispid;

	if( code == FKT_NANOSLEEP_ENTRY_CODE  || code == FKT_NANOSLEEP_EXIT_CODE )
		{
		k = v_pos;
		i = 0;
		fprintf(stdbug, "v[%d]: %s, code=%04x, counter=%d, "
						"cycles_array[%d]=%llu, caller_pid[%d]=%d\n",
						k, find_name(v[k].code,1), v[k].code, v[k].counter,
						i, v[k].cycles_array[i], i, v[k].caller_pid[i]);
		}

	v_pos++;
	}

/*	draw a horizontal line of 80 c characters onto stdout */
void my_print_line( int c )
	{
	int		i;

	for( i = 0;  i < 80;  i++ )
		putchar(c);
	putchar('\n');
	}


void draw_histogram( int n, double histo_max, char **histo_name,
							u_64 *histo_cycles, double *histo_percent)
	{
	int		i, k, length;
	u_64	sum;

	printf("\n");
	printf("%22s  %4s %12s %8s\n\n", "Name", "Code", "Cycles", "Percent");
	sum = 0LL;
	for( k = 0;  k < n;  k++ )
		{
		printf("%28s %12llu %7.2f%% ",
							histo_name[k], histo_cycles[k], histo_percent[k]);
		length = (int)((histo_percent[k]*MAX_HISTO_BAR_LENGTH)/histo_max + 0.5);
		for( i = 0;  i < length;  i++ )
			putchar(HISTO_CHAR);
		putchar('\n');
		sum += histo_cycles[k];
		}
	printf("%22s  %4s %12llu %7.2f%%\n\n",
										"Total elapsed cycles", "", sum, 100.0);
	}


void my_print_fun_helper(u_64 CycPerJiffy)
	{
	int						i, z, k;
	u_64					sum_other_cycles = 0LL;
	u_64					delta;
	u_64					timediff, totaltimetaken, sum;
	/*****
	double					Timems;
	*****/
	double					Main_fun_sum, Main_fun_sum_percent;
	double					called_fun_percent, only_fun_percent;
	struct v3_stack_item	*v3_ptr;
	double					histo_max;
	char					*category, *name;
	char					*histo_name[HISTO_SIZE];
	u_64					histo_cycles[HISTO_SIZE];
	double					histo_percent[HISTO_SIZE];


	printf("\n\n");
	printf("%30s ", "");
	for( i = 0;  i < ncpus;  i++ )
		printf("%10s%2d", "cpu", i);
	printf("\n");

	printf ("%30s:", "Cycles at start of stat taking");
	for( i = 0;  i < ncpus;  i++ )
		printf("%12llu", start_time_cpu[i]);
	printf("\n");

	printf ("%30s:", "Cycles at end of stat taking");
	for( i = 0;  i < ncpus;  i++ )
		printf("%12llu", end_time_cpu[i]);
	printf("\n");

	totaltimetaken = 0ll;
	printf ("%30s:", "Number of elapsed stat cycles");
	for( i = 0;  i < ncpus;  i++ )
		{
		timediff = end_time_cpu[i] - start_time_cpu[i];
		totaltimetaken += timediff;
		printf("%12llu", timediff);
		}
	printf("\n\n");
	printf ("%30s:%12llu\n", "Total elapsed stat cycles", totaltimetaken);

	/*****
	printf ("%36s : %10llu\n", "Cycles Per Jiffy", CycPerJiffy);
	Timems = ((double)TimeDiff) * 10.0 / ((double)CycPerJiffy);
	printf ("%36s : %10.2fms\n", "Time taken is", Timems);
	*****/

	printf("\n\n");

/*	print table of all function names */
if( v_pos > 0 )
	{
	printf("%30s  %4s %12s %10s %11s\n",
			"Name", "Code", "Cycles", "Count", "Avg");
	printf("%30s  %4s %12s %10s %11s\n",
			"----", "----", "------", "-----", "---");
	for ( k = 0;  k < v_pos;  k++ )
		{
		if( v[k].code > 0 )
			{
			for ( z = 0; z < MAX_EACH_FUN_RECUR; z++)
				{
				if (v[k].cycles_array[z] != 0LL)
					{/* unterminated call to this function */
    				u_64					cycdiff;
					struct v3_stack_item	*v3_ptr;

					/* find the v3 stack item for pid of caller to this func */
					for( v3_ptr=v3_stack_head; v3_ptr; v3_ptr=v3_ptr->v3_next )
						{
						if( v3_ptr->v3_pid == v[k].caller_pid[z] )
							break;
						}
					if( v3_ptr == NULL )
						{/* could not find v3 stack item for this caller ?! */
						printf("%30s  %04x %12u ** unexited call, "
								"bad pid %d\n",
								find_name(v[k].code,0), v[k].code, 0,
								v[k].caller_pid[z]);
						}
					else
						{/* found an existing stack for the caller's pid */

						/* compute time spent in last call to this function */
						if(v3_ptr->v3_code[v3_ptr->v3_pos]==FKT_SWITCH_TO_CODE)
							{/* top of pid's stack is switch_to */
							cycdiff = v3_ptr->v3_start_cycle[v3_ptr->v3_pos] -
														v[k].cycles_array[z];
							}
						else if (v3_ptr->v3_code[v3_ptr->v3_pos]==v[k].code )
							{/* top of pid's stack is call to this func */
							cycdiff = 0LL;
							}
						else
							{/* top of current stack not switch_to, should be */
							fprintf(stderr,
							"=== top of stack for pid %u not switch_to\n",
									v3_ptr->v3_pid);
							fprintf(stdout,
							"=== top of stack for pid %u not switch_to\n",
									v3_ptr->v3_pid);
							cycdiff = 0LL;
							}

						printf("%30s  %04x %12llu ** unexited call, "
								"pid %d\n",
								find_name(v[k].code,0), v[k].code, cycdiff,
								v[k].caller_pid[z]);
						/* add this time to total time ever spent in function */
						if( v[k].code == FKT_NANOSLEEP_ENTRY_CODE )
							fprintf(stdbug,
							"add %llu to nanosleep_entry v[%d].total %llu\n",
													cycdiff, k, v[k].total);
						if( v[k].code == FKT_NANOSLEEP_EXIT_CODE )
							fprintf(stdbug,
							"add %llu to nanosleep_exit v[%d].total %llu\n",
													cycdiff, k, v[k].total);
                		v[k].total += cycdiff;
						}
					}
				}
			printf("%30s  %04x %12llu %10d", 
				   find_name(v[k].code,0), v[k].code, v[k].total, v[k].counter);
			if( v[k].counter > 0 )
				printf(" %11.2f", (double)v[k].total/v[k].counter);
			printf("\n");
			}
		}

	printf("\n\n");
	printf("%4s %25s  %4s %12s %10s\n",
			"From", "To", "Code", "Cycles", "Count");
	printf("%4s %25s  %4s %12s %10s\n",
			"----", "--", "----", "------", "-----");
	for ( k = 0;  k < v_pos;  k++ )
		{
		if( v[k].code > 0 )
			{
			printf("%s\n", find_name(v[k].code,0));
			for ( z = 0;  z < v_pos;  z++ )
				{
				if ( v[k].called_funs_ctr[z] != 0 ) 
					{
					printf("%30s  %04x %12llu %10d\n", 
							find_name(v[z].code,0), v[z].code,
							v[k].called_funs_cycles[z], 
							 v[k].called_funs_ctr[z]);
					}  
				}
			printf("\n");
			}
		}
	printf("\n\n");

	printf("%50s\n\n", "NESTING SUMMARY");
	my_print_line('*');
	printf("\n");
	for ( k = 0;  k < v_pos;  k++ )
		{
		if( v[k].code > 0 )
			{
			printf("%32s  %4s %12s %7s %11s %8s\n\n",
					"Name", "Code", "Cycles", "Count", "Average", "Percent");
			printf("%32s  %4s %12s %7s %11s %8s\n\n",
					"----", "----", "------", "-----", "-------", "-------");
			Main_fun_sum = (double)v[k].total;
			Main_fun_sum_percent = 100.0;
			printf("%32s  %04x %12llu %7d %11.2f %7.2f%%\n\n", 
				   				find_name(v[k].code,0), v[k].code, v[k].total,
				   				v[k].counter, (double)v[k].total/v[k].counter, 
								Main_fun_sum_percent);

			for ( z = 0;  z < v_pos;  z++ )
				{
				if( v[z].code > 0  &&  v[k].called_funs_ctr[z] != 0 ) 
					{
					called_fun_percent = (double)v[k].called_funs_cycles[z] /
														Main_fun_sum * 100.0;
					printf("%32s  %04x %12llu %7d %11.2f %7.2f%%\n", 
										find_name(v[z].code,0), v[z].code,
										v[k].called_funs_cycles[z], 
										v[k].called_funs_ctr[z], 
										(double) v[k].called_funs_cycles[z] /
										v[k].called_funs_ctr[z],
										called_fun_percent);
					sum_other_cycles += v[k].called_funs_cycles[z];
					}
				}

			printf("%27s ONLY  %04x ", find_name(v[k].code,0), v[k].code);

			if( sum_other_cycles > v[k].total )
				printf("%12llu ** sum of cycles too big **\n",sum_other_cycles);
			else
				{
				only_fun_percent = (double)(v[k].total - sum_other_cycles) /
														Main_fun_sum * 100.0;
				printf("%12llu %7d %11.2f %7.2f%%\n", 
							v[k].total - sum_other_cycles,
							v[k].counter,
							(double)(v[k].total-sum_other_cycles)/v[k].counter,
							only_fun_percent); 
				}
			sum_other_cycles = 0LL;
			my_print_line('-');
			printf("\n");
			}
		}
	}

	printf("\n\n");
	/*****
	printf("%37s: %12llu cycles %9.3f ms %7.2f%%\n", "Total time taken",
													TimeDiff, Timems, 100.0);
	*****/

	/* now print out the trap/syscall/irq statistics from the syscall arrays */
	printf("\n\n\n");
	printf("%8s %23s  %4s %12s %7s %11s\n",
	"", "Name", "Code", "Cycles", "Count", "Average");
	printf("%8s %23s  %4s %12s %7s %11s\n",
	"", "----", "----", "------", "-----", "-------");
	for( i = 0;  i < FKT_UNSHIFTED_LIMIT_CODE; i++ )
		{
		if( syscall_count[i] > 0 )
			{/* this occurred at least once, print it */
			int		should_be_hex;

			z = find_syscall(i, &category, &name, &should_be_hex);
			printf("%8s %23s  ", category, name);
			if( should_be_hex )
				printf("%04x ", z);
			else
				printf("%4d ", z);
			printf("%12llu %7u %11.2f\n",
			 			syscall_cycles[i], syscall_count[i],
						((double)syscall_cycles[i])/((double)syscall_count[i]));
			}
		}


	/*	close up all open stacks */
	/*	stacks are typically left open when a process is blocked in the kernel*/
	for( v3_ptr = v3_stack_head;  v3_ptr != NULL;  v3_ptr = v3_ptr->v3_next )
		{
		if( v3_ptr->v3_pos > 1 )
			{
			fprintf(stdbug, "=== final stack for pid %u has pos %d\n",
											v3_ptr->v3_pid, v3_ptr->v3_pos);
			}
		if( v3_ptr->v3_pos > 0  &&
						v3_ptr->v3_code[v3_ptr->v3_pos] != FKT_SWITCH_TO_CODE )
			{/* last thing process did in kernel was NOT a block!!! */
			fprintf(stderr, "=== top of final stack for pid %u not switch_to\n",
																v3_ptr->v3_pid);
			fprintf(stdout, "=== top of final stack for pid %u not switch_to\n",
																v3_ptr->v3_pid);
			}
		else if( v3_ptr->v3_pos > 0 )
			{
			while( v3_ptr->v3_pos > 1 )
				{/* close up all unterminated function calls */
				unsigned int	code;

				code = v3_ptr->v3_code[v3_ptr->v3_pos-1];
				delta = v3_ptr->v3_start_cycle[v3_ptr->v3_pos] -
						v3_ptr->v3_start_cycle[v3_ptr->v3_pos-1];
    			if( code < FKT_UNSHIFTED_LIMIT_CODE )
    				{/* this is a trap/syscall/irq entry */
					v3_syscall_cycles[code] +=  delta;
					v3_syscall_count[code] += 1;
					}
				else
					{/* this is a function entry */
					update_function(delta, code);
					}
				v3_ptr->v3_pos -= 1;
				}
			v3_ptr->v3_process_cycles = v3_ptr->v3_start_cycle[1] -
										v3_ptr->v3_start_cycle[0];
			}
		}

	/* now print out the precise accounting statistics from the v3 arrays */
	sum = 0LL;
	k = 0;
	histo_max = 0.0;
	printf("\n\n\n");
	printf("%60s\n\n", "Accounting for every cycle exactly once");
	my_print_line('*');
	printf("%8s %23s  %4s %12s %7s %11s %8s\n",
	"", "Name", "Code", "Cycles", "Count", "Average", "Percent");
	my_print_line('*');
	for( i = 0;  i < FKT_UNSHIFTED_LIMIT_CODE; i++ )
		{
		if( v3_syscall_count[i] > 0 )
			{/* this occurred at least once, print it */
			int		should_be_hex;

			z = find_syscall(i, &category, &name, &should_be_hex);
			if( (histo_name[k] = malloc(strlen(name)+8)) == NULL )
				{
				fprintf(stderr, "no space to malloc histo_name\n");
				exit(EXIT_FAILURE);
				}
			histo_cycles[k] = v3_syscall_cycles[i];
			sum += histo_cycles[k];
			histo_percent[k] = 
						100.0*((double)histo_cycles[k])/(double)totaltimetaken;
			if( histo_max < histo_percent[k] )
				histo_max = histo_percent[k];
			printf("%8s %23s  ", category, name);
			if( should_be_hex )
				{
				printf("%04x ", z);
				sprintf(histo_name[k], "%s  %04x", name, z);
				}
			else
				{
				printf("%4d ", z);
				sprintf(histo_name[k], "%s  %4d", name, z);
				}
			printf("%12llu %7u %11.2f %7.2f%%\n",
				histo_cycles[k], v3_syscall_count[i],
				((double)histo_cycles[k])/((double)v3_syscall_count[i]),
				histo_percent[k]);
			k++;
			}
		}
	for( i = 0;  i < MAX_FUNS; i++ )
		{
		z = v3_function_code[i];
		if( z > 0  &&  z != FKT_SWITCH_TO_CODE )
			{/* this occurred at least once and was not switch_to, print it */

			name = find_name(z,0);
			if( (histo_name[k] = malloc(strlen(name)+8)) == NULL )
				{
				fprintf(stderr, "no space to malloc histo_name\n");
				exit(EXIT_FAILURE);
				}
			sprintf(histo_name[k], "%s  %04x", name, z);
			histo_cycles[k] = v3_function_cycles[i];
			sum += histo_cycles[k];
			histo_percent[k] = 
						100.0*((double)histo_cycles[k])/(double)totaltimetaken;
			if( histo_max < histo_percent[k] )
				histo_max = histo_percent[k];
			printf("%8s %23s", "function", name);
			printf("  %04x %12llu %7u %11.2f %7.2f%%\n",
					z, histo_cycles[k], v3_function_count[i],
					((double)histo_cycles[k])/((double)v3_function_count[i]),
					histo_percent[k]);
			k++;
			}
		}


	/*	now print out process information from stacks */
	for( v3_ptr = v3_stack_head;  v3_ptr != NULL;  v3_ptr = v3_ptr->v3_next )
		{
		if( v3_ptr == v3_stack_first )
			{/* this is the user process */
			z = v3_stack_first->v3_pid;
			name = "user process";
			category = "user";
			/* histo_cycles[k] = v3_stack_first->v3_start_cycle[0]; */
			histo_cycles[k] = v3_stack_first->v3_process_cycles;
			}
		else if( (z = v3_ptr->v3_pid) == 0 )
			{/* this is the idle process (z == 0) */
			name = "idle process";
			category = "idle";
			histo_cycles[k] = v3_ptr->v3_process_cycles;
			}
		else
			{/* this is some other process (z != 0) */
			name = "other process";
			category = "other";
			histo_cycles[k] = v3_ptr->v3_process_cycles;
			}

		if( (histo_name[k] = malloc(strlen(name)+8)) == NULL )
			{
			fprintf(stderr, "no space to malloc histo_name\n");
			exit(EXIT_FAILURE);
			}
		sprintf(histo_name[k], "%s  %4u", name, z);
		sum += histo_cycles[k];
		histo_percent[k] = 
					100.0*((double)histo_cycles[k])/(double)totaltimetaken;
		if( histo_max < histo_percent[k] )
			histo_max = histo_percent[k];
		printf("%8s %23s", category, name);
		printf("  %4u %12llu %7s %11s %7.2f%%\n",
							z, histo_cycles[k], "", "",
							histo_percent[k]);
		k++;
		}
	printf("%8s %23s", "Total", "Total elapsed cycles");
	printf("  %4s %12llu %7s %11s %7.2f%%\n", "", sum, "", "",
									100.0*((double)sum)/(double)totaltimetaken);

	printf("\n\n%65s\n", "Histogram accounting for every cycle exactly once");
	draw_histogram(k, histo_max, histo_name, histo_cycles, histo_percent);
	}


/*	called once to start statistics taking */
void start_taking_stats( unsigned int thispid, u_64 reltime )
	{
	fprintf(stdbug, "start taking stats, lastcpu %u, pid %u, rel time %llu\n",
													lastcpu, thispid, reltime);
	fun_time = 1;							/* start the statistics taking */
	start_time_glob = reltime;				/* remember the starting time */
	if( v3_stack_ptr[lastcpu] != NULL )
		{
		fprintf(stderr, "=== starting v3_stack_ptr[%u] = %p\n",
												lastcpu, v3_stack_ptr[lastcpu]);
		fprintf(stdout, "=== starting v3_stack_ptr[%u] = %p\n",
												lastcpu, v3_stack_ptr[lastcpu]);
		exit(EXIT_FAILURE);
		}
	if( v3_stack_head != NULL )
		{
		fprintf(stderr, "=== starting v3_stack_head = %p\n",
														v3_stack_head);
		fprintf(stdout, "=== starting v3_stack_head = %p\n",
														v3_stack_head);
		exit(EXIT_FAILURE);
		}
	v3_stack_ptr[lastcpu] = create_stack(thispid, lastcpu, reltime);
	}


/*	called once to stop statistics taking */
void stop_taking_stats( unsigned int thispid, u_64 reltime )
	{/* this times() occurred while taking statistics */
	fprintf(stdbug, "stop taking stats, pid %u, rel time %llu\n",
															thispid, reltime);
	fun_time = -1;				/* stop the statistics taking */
	end_time_glob = reltime;	/* remember the stopping time */
	if( v3_stack_ptr[lastcpu] == NULL )
		{
		fprintf(stderr,"=== pid %4u stopping v3_stack_ptr[%d] = %p\n",
									thispid, lastcpu, v3_stack_ptr[lastcpu]);
		fprintf(stdout,"=== pid %4u stopping v3_stack_ptr[%d] = %p\n",
									thispid, lastcpu, v3_stack_ptr[lastcpu]);
		exit(EXIT_FAILURE);
		}
	if( v3_stack_ptr[lastcpu]->v3_pos > 0 )
		{
		fprintf(stderr, "=== pid %4u stopping v3_pos = %u\n",
										thispid, v3_stack_ptr[lastcpu]->v3_pos);
		fprintf(stdout, "=== pid %4u stopping v3_pos = %u\n",
										thispid, v3_stack_ptr[lastcpu]->v3_pos);
		close_stack(v3_stack_ptr[lastcpu], thispid, reltime);
		}
	v3_stack_ptr[lastcpu]->v3_process_cycles =
							reltime - v3_stack_ptr[lastcpu]->v3_start_cycle[0];
	}


/*	make sure this cpu is in the correct statistics taking state */
void adjust_cpu_state( unsigned int thiscpu, u_64 reltime )
	{
	if( fun_time != fun_time_cpu[thiscpu] )
		{/* first time this cpu is active since a stat-taking state change */
		if( fun_time_cpu[thiscpu] == 0  ||  fun_time > 0 )
			{/* this cpu not active yet, must get a start time for it */
			start_time_cpu[thiscpu] = reltime;
			fprintf(stdbug, "cpu %u start time %llu\n", thiscpu, reltime);
			}
		if( fun_time < 0 )
			{/* this cpu still active but stat-taking has stopped */
			end_time_cpu[thiscpu] = reltime;
			fprintf(stdbug, "cpu %u end time %llu\n", thiscpu, reltime);
			}
		/*	force cpu state to be the same as the global state */
		fun_time_cpu[thiscpu] = fun_time;
		}
	}


/*
bufptr[0] is the first item for each line ie cycles
bufptr[1] is pid in low half, processor number in high half
bufptr[2] is code
bufptr[3] is P1
bufptr[4] is P2
...
bufptr[n] is P(n-2)
*/

/*	called to dump the next record from the buffer */
unsigned int *dumpslot( unsigned int *n,  unsigned int *bufptr )
	{
	unsigned int	code, fullcode, params, i, thispid, thiscpu;
	u_64			reltime, r, ptime, x;
	int				print_this = 0;
	char			workspace[128];


	fprintf(stdbug,"\n");

	thiscpu = bufptr[1] >> 16;				/* get cpu number */

	if( thiscpu > ncpus )
		{
		fprintf(stderr, "thiscpu = %u, limit is %u\n", thiscpu, ncpus);
		fprintf(stdout, "thiscpu = %u, limit is %u\n", thiscpu, ncpus);
		}
	else if( !already_saw[thiscpu] )
		{/* this is first time we have seen this cpu */
		lasttime[thiscpu] = basetime[thiscpu] = (u_64)bufptr[0];
		fprintf(stdout, "initial: basetime[%u] %llu\n",
											thiscpu, basetime[thiscpu]);
		already_saw[thiscpu] = 1;			/* mark that we have seen this cpu*/
		lastcpu = thiscpu;
		}
	else if( lastcpu != thiscpu )
		{/* changing cpus */
		fprintf(stdbug, "changing from cpu %u to cpu %u\n", lastcpu, thiscpu);
		lastcpu = thiscpu;
		}

	x = ((u_64)bufptr[0])+basehigh[thiscpu];/* cnv 32-bit cycles to 64-bit */
	if( x < lasttime[thiscpu] )
		{
		fprintf(stdbug,"cycle overflow: x %llu, lasttime %llu, basehigh %llu\n",
									x, lasttime[thiscpu], basehigh[thiscpu]);
		basehigh[thiscpu]+= 0x100000000LL;	/* add 1 to high 32-bits */
		x				 += 0x100000000LL;	/* add 1 to high 32-bits */
		fprintf(stdbug,"after overflow: x %llu, lasttime %llu, basehigh %llu\n",
									x, lasttime[thiscpu], basehigh[thiscpu]);
		}
	lasttime[thiscpu] = x;
	reltime = x - basetime[thiscpu];
	r = reltime-lastreltime[thiscpu];
	if( print_delta_time )
		ptime = r;					/* print time delta from last entry */
	else
		ptime = reltime;			/* print cummulative time on cpu */
	fullcode = bufptr[2];
	code = fullcode >> 8;
	if( code == FKT_CALIBRATE0_CODE  ||
		code == FKT_CALIBRATE1_CODE  ||
		code == FKT_CALIBRATE2_CODE )
		print_this = 1;
	thispid = bufptr[1]&0xffff;
    if (thispid == kidpid)
		{
        kidflag = 1;
		}
	fprintf(stdbug,
	"dumpslot: thiscpu %u, lastpid %d, thispid %u, code 0x%06x, delta %llu\n",
						thiscpu, lastpid[thiscpu], thispid, fullcode, r);
	if( all_stats )
		{
		print_this = 1;
		if( code == FKT_SWITCH_TO_CODE  &&  fun_time == 0 )
			start_taking_stats(thispid, reltime);		/* start taking stats */
		else if( fullcode == __NR_fkt_keychange  &&  fun_time > 0 )
			stop_taking_stats(thispid, reltime);		/* stop taking stats */
		/* else statistics were already finished, do nothing more */
		}
    else if( (thispid == kidpid) && (fullcode == __NR_times) )
		{/* this is a call to times() by the kid task */
		if( fun_time == 0 )
			{/* this is the first times() call for this kid task */
			start_taking_stats(thispid, reltime);		/* start taking stats */
			}
		else if( fun_time > 0 )
			{/* this times() occurred while taking statistics */
			stop_taking_stats(thispid, reltime);		/* stop taking stats */
			}
		/* else statistics were already finished, do nothing more */
		}
    
	/*	make sure this cpu is in the correct statistics taking state */
	adjust_cpu_state(thiscpu, reltime);

	if( fun_time > 0 )
		update_pid_cycles(r, thispid);

    if (kidflag|print_this)
		printf( "%10llu%4u %4u", ptime, thiscpu, thispid);

	if( fullcode < FKT_UNSHIFTED_LIMIT_CODE )
		{/* unshifted code, no parameters */
		code = fullcode;
    	if (kidflag|print_this)
			printf(" %04x", fullcode);
		if( fullcode < FKT_SYS_CALL_LIMIT_CODE )
			{
			i = fullcode;
			sprintf(workspace, "system call %u", i);
			if( i > NSYS_CALLS )
				i = NSYS_CALLS;
	        if (kidflag|print_this)
				printf(" %28s  %-s", workspace, sys_calls[i]);
			}
		else if( fullcode < FKT_TRAP_LIMIT_CODE )
			{
			i = fullcode - FKT_SYS_CALL_LIMIT_CODE;
			sprintf(workspace, "trap %u", i);
			if( i > NTRAPS )
				i = NTRAPS;
        	if (kidflag|print_this)
				printf(" %28s  %-s", workspace, traps[i]);
			}
		else
			{
			i = fullcode - FKT_TRAP_LIMIT_CODE;
			sprintf(workspace, "IRQ %u", i);
        	if (kidflag|print_this)
				printf(" %28s  %-s", workspace, find_irq(i));
			}
		params = 3;
		}
	else
		{/* shifted code */
    	if (kidflag|print_this)
			printf(" %04x", code);
		params = bufptr[2] & 0xff;
		if( params & 0x03 )
			{
			printf(
				"***** header size not multiple of 4: %08x, %04x, %d *****\n",
													bufptr[2], code, params);
			exit(EXIT_FAILURE);
			}
		params = (params>>2);
        if (kidflag|print_this)
			printf(" %28s",  find_name(code, 1));
		}
    if (kidflag|print_this)
		{
		for( i = 3;  i < params;  i++ )
			{
			if( i == 3 )
				printf("  ");				/* first time only */
			else
				printf(", ");				/* not the first time */
			if( bufptr[i] > 0xffffff00 )	/* -256 */
				printf("%d", (int)bufptr[i]);
			else if( bufptr[i] <= 0xfffffff )
				printf("%u", bufptr[i]);
			else
				printf("%#08x", bufptr[i]);
			}
		printf("\n");
		}
	*n += params;
    if( (kidflag || all_stats)  &&  fun_time > 0 )
		{
		if( params <= 4 )
        	my_print_fun(code, reltime, thispid, bufptr[3], 0);
		else
        	my_print_fun(code, reltime, thispid, bufptr[3], bufptr[4]);
		}

	lastreltime[thiscpu] = reltime;
	lastpid[thiscpu] = thispid;

	return &bufptr[params];
	}

/*	returns 0 if ok, else -1 */
int dumpit( unsigned int *buffer, unsigned int nints )
	{
	unsigned int			n;
	unsigned int			*bufptr, *buflimit;
	u_64					ratio;
	struct v3_stack_item	*v3_ptr;


	buflimit = buffer + nints;

	printf( "%10s%4s %4s %4s %28s", "cycles", "cpu", "pid", "code", "name");
	printf("  P1, P2, P3, ...\n");

	for( bufptr = buffer, n = 0;  n < nints && bufptr < buflimit; )
		{
		bufptr = dumpslot(&n, bufptr);
		}
	if( fun_time > 0 )				/* never stopped stat taking */
		stop_taking_stats(lastpid[lastcpu], lastreltime[lastcpu]);

	/* end times on all cpus */
	for( n = 0;  n < ncpus;  n++ )
		adjust_cpu_state(n, lastreltime[n]);

	/* now close up all stacks */
	for(v3_ptr = v3_stack_head;
		v3_ptr != NULL;
		v3_ptr = v3_ptr->v3_next )
			{
			if( v3_ptr->v3_pos > 0 )
				close_stack(v3_ptr,v3_ptr->v3_pid,lastreltime[v3_ptr->v3_cpu]);
			}

	printf("\n");
	ratio = (u_64)(stop_jiffies - start_jiffies);
	if( ratio != 0LL )
		ratio = lastreltime[lastcpu] / ratio;
	printf("%llu clock cycles in %u jiffies = %llu cycles per jiffy\n",
			lastreltime[lastcpu], (unsigned int)(stop_jiffies - start_jiffies),
			ratio);
	printf("\n");
	print_cycles(ratio);
	my_print_fun_helper(ratio);
	return 0;
	}


void get_the_time( time_t *the_time, char *message )
	{
	struct tm	breakout;
	char		buffer[128];


	if( read(fd, (void *)the_time, sizeof(time_t)) <= 0 )
		{
		perror(message);
		exit(EXIT_FAILURE);
		}
	if( localtime_r(the_time, &breakout) == NULL )
		{
		fprintf(stderr, "Unable to break out %s\n", message);
		exit(EXIT_FAILURE);
		}
	if( strftime(buffer, 128, "%d-%b-%Y %H:%M:%S", &breakout) == 0 )
		{
		fprintf(stderr, "Unable to convert %s\n", message);
		exit(EXIT_FAILURE);
		}
	printf("%14s = %s\n", message, buffer);
	}


int main( int argc, char *argv[] )
	{
	unsigned int	nints, max_nints;
	unsigned int	*bufptr;
	char			*infile;
	int				c, i;
	double			*mhz;


	kidpid = 0;
	infile = NULL;
	stdbug = NULL;
	opterr = 0;

	while( (c = getopt(argc, argv, OPTIONS)) != EOF )
		switch( c )
			{
		case 'a':
			all_stats = 1;
			break;
		case 'd':
			stdbug = stdout;
			break;
		case 'c':
			print_delta_time = 0;
			break;
		case 'p':
			kidflag = 1;
			break;
		case 'f':
			infile = optarg;
			break;
		case ':':
			fprintf(stderr, "missing parameter to switch %c\n", optopt);
			exit(EXIT_FAILURE);
		case '?':
			fprintf(stderr, "illegal switch %c\n", optopt);
			exit(EXIT_FAILURE);
			}	/* switch */

	if( optind < argc )
		fprintf(stderr, "extra command line arguments ignored\n");

	if( stdbug == NULL )
		{
		if( (stdbug = fopen("/dev/null", "w")) == NULL )
			{
			perror("/dev/null");
			stdbug = stdout;
			}
		}
		
	if( infile == NULL )
		{
		if( (infile = getenv("TRACE_FILE")) == NULL )
			infile = DEFAULT_TRACE_FILE;
		}

	while( (fd = open(infile, O_RDONLY)) < 0 )
		{/* could not open the indicated file name for reading */
		perror(infile);
		if( strcmp(infile, DEFAULT_TRACE_FILE) == 0 )
			/* can't open the default file, have to give up */
			exit(EXIT_FAILURE);
		infile = DEFAULT_TRACE_FILE;
		fprintf(stderr, "input defaulting to %s\n", infile);
		}

	/*	the order of reading items from the trace file must obviously
		correspond with the order they were written by fkt_record	*/
	if( read(fd, (void *)&ncpus, sizeof(ncpus)) < 0 )
		{
		perror("read ncpus");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %i\n", "ncpus", ncpus);

	if( (mhz = (double *)malloc(ncpus * sizeof(double))) == NULL )
		{
		fprintf(stderr, "unable to malloc mhz[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	for( i = 0;  i < ncpus;  i++ )
		{
		if( read(fd, (void *)&mhz[i], sizeof(mhz[i])) < 0 )
			{
			perror("read mhz[i]");
			exit(EXIT_FAILURE);
			}
		printf("%12s %d = %.3f MHZ\n", "cpu", i, mhz[i]);
		}

	if( (basetime = (u_64 *)calloc(ncpus, sizeof(u_64))) == NULL )
		{
		fprintf(stderr, "unable to malloc basetime[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (basehigh = (u_64 *)calloc(ncpus, sizeof(u_64))) == NULL )
		{
		fprintf(stderr, "unable to malloc basehigh[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (lasttime = (u_64 *)calloc(ncpus, sizeof(u_64))) == NULL )
		{
		fprintf(stderr, "unable to malloc lasttime[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (lastreltime = (u_64 *)calloc(ncpus, sizeof(u_64))) == NULL )
		{
		fprintf(stderr, "unable to malloc lastreltime[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (start_time_cpu = (u_64 *)calloc(ncpus, sizeof(u_64))) == NULL )
		{
		fprintf(stderr, "unable to malloc start_time_cpu[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (end_time_cpu = (u_64 *)calloc(ncpus, sizeof(u_64))) == NULL )
		{
		fprintf(stderr, "unable to malloc end_time_cpu[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (lastpid=(unsigned int *)calloc(ncpus,sizeof(unsigned int))) == NULL )
		{
		fprintf(stderr, "unable to malloc lastpid[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (already_saw = (int *)calloc(ncpus, sizeof(int))) == NULL )
		{
		fprintf(stderr, "unable to malloc already_saw[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (fun_time_cpu = (int *)calloc(ncpus, sizeof(int))) == NULL )
		{
		fprintf(stderr, "unable to malloc fun_time_cpu[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}
	if( (v3_stack_ptr=(struct v3_stack_item **)calloc(ncpus,sizeof(void *)))
																	== NULL )
		{
		fprintf(stderr, "unable to malloc v3_stack_ptr[%d]\n", ncpus);
		exit(EXIT_FAILURE);
		}

	if( read(fd, (void *)&pid, sizeof(pid)) <= 0 )
		{
		perror("read pid");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %lu\n", "pid", pid);

	if( read(fd, (void *)&kidpid, sizeof(kidpid)) <= 0 )
		{
		perror("read kidpid");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %lu\n", "kidpid", kidpid);


	get_the_time(&start_time, "start time");
	get_the_time(&stop_time, "stop time");

	if( read(fd, (void *)&start_jiffies, sizeof(start_jiffies)) <= 0 )
		{
		perror("read start_jiffies");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %lu\n", "start jiffies", (unsigned long)start_jiffies);

	if( read(fd, (void *)&stop_jiffies, sizeof(stop_jiffies)) <= 0 )
		{
		perror("read stop_jiffies");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %lu\n", "stop jiffies", (unsigned long)stop_jiffies);
	printf("%14s = %lu\n", "total jiffies", (unsigned long)stop_jiffies -
												 (unsigned long)start_jiffies);

	if( read(fd, (void *)&nirqs, sizeof(nirqs)) <= 0 )
		{
		perror("read nirqs");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %u\n", "nirqs", nirqs);
	if( read(fd, (void *)&ncpus, sizeof(ncpus)) <= 0 )
		{
		perror("read ncpus");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %u\n", "ncpus", ncpus);
	load_irqs();

	if( read(fd, (void *)&max_nints, sizeof(max_nints)) <= 0 )
		{
		perror("read max_nints");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %u\n", "max_nints", max_nints);
	if( read(fd, (void *)&nints, sizeof(nints)) <= 0 )
		{
		perror("read nints");
		exit(EXIT_FAILURE);
		}
	printf("%14s = %u\n", "nints", nints);
	printf("\n");

	if( (bufptr = (unsigned int *)malloc(nints*4)) == NULL )
		{
		fprintf(stderr, "unable to malloc %u ints\n", nints);
		exit(EXIT_FAILURE);
		}

	if( read(fd, (void *)bufptr, nints*4) <= 0 )
		{
		perror("read buffer");
		exit(EXIT_FAILURE);
		}

	for( i = 0;  i < MAX_FUNS;  i++ )
		{
		v3_function_cycles[i] = 0;
		v3_function_count[i] = 0;
		v3_function_code[i] = 0;
		}

	/*	clear out the totals for trap/syscall/irq measurements */
	for( i = 0;  i < FKT_UNSHIFTED_LIMIT_CODE; i++ )
		{
		syscall_cycles[i] = 0;
		syscall_count[i] = 0;
		v3_syscall_cycles[i] = 0;
		v3_syscall_count[i] = 0;
		}

	dumpit(bufptr, nints);

	/* print final statistics about storage usage */
	printf("\n\n");
	printf("internal storage usage statistics\n\n");

	printf("%21s %3u, last possible %3d, percentage used %7.2f%%\n\n",
			"last used nints",
			nints, max_nints,
			((double)nints)*100.0/((double)(max_nints)));
	printf("%21s %3u, last possible %3d, percentage used %7.2f%%\n",
			"last used v2_pos",
			actual_max_v2_pos, MAX_UNSHIFTED_FUN_RECUR-1,
			((double)actual_max_v2_pos)*100.0/
										((double)(MAX_UNSHIFTED_FUN_RECUR-1)));
	printf("%21s %3u, last possible %3d, percentage used %7.2f%%\n",
			"last used v3_pos",
			actual_max_v3_pos, MAX_SHIFTED_FUN_RECUR-1,
			((double)actual_max_v3_pos)*100.0/
										((double)(MAX_SHIFTED_FUN_RECUR-1)));
	printf("%21s %3u, last possible %3d, percentage used %7.2f%%\n",
			"last used cycles_pos",
			actual_max_cycles_pos, MAX_EACH_FUN_RECUR-1,
			((double)actual_max_cycles_pos)*100.0/
											((double)(MAX_EACH_FUN_RECUR-1)));
	printf("%21s %3u, last possible %3d, percentage used %7.2f%%\n",
			"last used v_pos",
			actual_max_v_pos, MAX_FUNS-1,
			((double)actual_max_v_pos)*100.0/((double)(MAX_FUNS-1)));
	printf("%21s %3u, last possible %3d, percentage used %7.2f%%\n",
			"last used v3_function",
			actual_max_v3_function, MAX_FUNS-1,
			((double)actual_max_v3_function)*100.0/((double)(MAX_FUNS-1)));
	return 0;
	}
