/*	fkt_record.c

	fkt_record --	control program to allow a user to trace his/her programs
					with fkt (fast kernel tracing) in the kernel.

	Copyright (C) 2000, 2001  Robert D. Russell -- rdr@unh.edu

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



	Usage: fkt_record [-k mask] [-f output-file] [program p1 p2 p3 ...]

			if -k is given, use the indicated mask to enable tracing.
			if that is not given, use mask of -1 (everything).

			if -f is not given, look for TRACE_FILE environment variable, and
			if that is not given, write to "trace_file".

			if program is not given, just spawn a nonsense kid to burn cpu.
*/

#define _POSIX_C_SOURCE 199309L

#include <stdio.h>
#include <stdlib.h>
#define __USE_XOPEN
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <signal.h>
#include <linux/fkt.h>
#include "fkt.h"
#include "timer.h"
#include "get_cpu_info.h"


#define CODE	0x20
#ifndef WHITE_SPACE
#define WHITE_SPACE	" \t\n\f\v\r"
#endif

#define OPTIONS ":f:k:"
#define DEFAULT_TRACE_FILE "trace_file"
#define MAXCPUS	16


/*	user area to hold copy of system trace buffer */
static unsigned int	*buffer = NULL;

static clock_t	start_jiffies = 0, stop_jiffies = 0;
static time_t	start_time = 0, stop_time = 0;
static int fd;
static unsigned long pid, kpid;
static int max_nints = 0;

struct code_list_item	{
						unsigned int			code;
						char					*name;
						struct code_list_item	*link;
						};

#define LINE_SIZE	128

/*	reads /proc/interrupts to find out the IRQ assignment on this machine */
/*	also figures out number of CPUs */
void record_irqs( void )
	{
	FILE					*table;
	struct code_list_item	*ptr, *irq_list = NULL;
	char					*lptr, *xptr, line[LINE_SIZE];
	unsigned int			n, i, nirqs, ncpus;


	nirqs = 0;
	ncpus = 0;
	if( (table = fopen("/proc/interrupts", "r")) == NULL )
		{
		fprintf(stderr, "unable to open /proc/interrupts\n");
		write(fd, (void *)&nirqs, sizeof(nirqs));
		write(fd, (void *)&ncpus, sizeof(ncpus));
		return;
		}

	/*	first line of file is headers of form CPUn */
	if( fgets(line, LINE_SIZE, table) == NULL )
		{
		fprintf(stderr, "unable to read /proc/interrupts\n");
		write(fd, (void *)&nirqs, sizeof(nirqs));
		write(fd, (void *)&ncpus, sizeof(ncpus));
		fclose(table);
		return;
		}
	
	for( lptr = line;  strtok(lptr, WHITE_SPACE) != NULL;  ncpus++ )
		lptr = NULL;
	printf("       : ncpus %5u\n", ncpus);

	/*	remaining lines of file are the IRQs */
	nirqs = 0;
	while( fgets(line, LINE_SIZE, table) != NULL )
		{/* build circular list with 1 item for each interrupt */
		n = strtoul(line, &lptr, 10);	/* scan off leading number */
		if( *lptr != ':' )				/* which must end with : */
			continue;					/* or it is not an IRQ number */
		ptr = (struct code_list_item *)malloc(sizeof(struct code_list_item));
		ptr->code = n;
		
		/*	scan over the interrupt counts, one per cpu, and controller type */
		xptr = lptr;
		for( lptr++, i=0;  i<=ncpus && (xptr = strtok(lptr, WHITE_SPACE))!=NULL;
																		i++ )
			lptr = NULL;

		/*	get to the leading character of the interrupt name */
		xptr += strlen(xptr) + 1;
		xptr += strspn(xptr, WHITE_SPACE);

		i = strlen(xptr);
		ptr->name = malloc(i);
		strncpy(ptr->name, xptr, i);
		ptr->name[i-1] = '\0';
		if( irq_list == NULL )
			ptr->link = ptr;
		else
			{
			ptr->link = irq_list->link;
			irq_list->link = ptr;
			}
		irq_list = ptr;
		printf(" irq %2u: %s\n", ptr->code, ptr->name);
		nirqs++;
		}
	printf("       : nirqs %4u\n", nirqs);
	printf("       : ncpus %4u\n", ncpus);
	fclose(table);
	write(fd, (void *)&nirqs, sizeof(nirqs));
	write(fd, (void *)&ncpus, sizeof(ncpus));
	if( (ptr = irq_list) != NULL )
		{/* make circular list into singly-linked, NULL-terminated list */
		irq_list = irq_list->link;
		ptr->link = NULL;
		}
	for( ptr = irq_list;  ptr != NULL;  ptr = ptr->link )
		{
		write(fd, (void *)&ptr->code, sizeof(ptr->code));
		i = strlen(ptr->name)+1;
		write(fd, (void *)&i, sizeof(i));
		i = (i + 3) & ~3;
		write(fd, ptr->name, i);
		}
	}

/*	returns 0 if ok, else -1 */
int recordit( int ncpus, double mhz[] )
	{
	int			i, nints;


	errno = 0;
	while( (nints = fkt_getcopy(0, NULL)) < 0  &&  errno == EAGAIN )
		errno = 0;
	if( nints < 0 )
		{
		perror("fkt_getcopy0");
		}

	printf("       : nints %5d\n", nints);

	if( (buffer = malloc(nints*4)) == NULL )
		{
		fprintf(stderr, "unable to malloc %d ints\n", nints);
		exit(EXIT_FAILURE);
		}

	errno = 0;
	while( (nints = fkt_getcopy(nints, buffer)) < 0  &&  errno == EAGAIN )
		errno = 0;
	if( nints < 0 )
		{
		perror("fkt_getcopy");
		}

	printf("       : nints %5d\n", nints);

	if( write(fd, (void *)&ncpus, sizeof(ncpus)) < 0 )
		perror("write ncpus");
	for( i = 0;  i < ncpus;  i++ )
		{
		if( write(fd, (void *)&mhz[i], sizeof(mhz[i])) < 0 )
			perror("write mhz");
		}
	if( write(fd, (void *)&pid, sizeof(pid)) < 0 )
		perror("write pid");
	if( write(fd, (void *)&kpid, sizeof(kpid)) < 0 )
		perror("write kpid");
	if( write(fd, (void *)&start_time, sizeof(start_time)) < 0 )
		perror("write start_time");
	if( write(fd, (void *)&stop_time, sizeof(stop_time)) < 0 )
		perror("write stop_time");
	if( write(fd, (void *)&start_jiffies, sizeof(start_jiffies)) < 0 )
		perror("write start_jiffies");
	if( write(fd, (void *)&stop_jiffies, sizeof(stop_jiffies)) < 0 )
		perror("write stop_jiffies");
	record_irqs();
	if( write(fd, (void *)&max_nints, sizeof(max_nints)) < 0 )
		perror("write max_nints");
	if( write(fd, (void *)&nints, sizeof(nints)) < 0 )
		perror("write nints");
	if( write(fd, (void *)buffer, nints*4) < 0 )
		perror("write buffer");

	return 0;
	}


void dumptime( char *message , time_t *the_time, clock_t *the_jiffies )
	{
	struct tms		cur_time;
	struct tm		breakout;
	char			buffer[128];

	if( (*the_time = time(NULL)) == -1 )
		perror("time");

	if( (*the_jiffies = times(&cur_time)) < 0 )
		perror("times");

	if( localtime_r(the_time, &breakout) == NULL )
		fprintf(stderr, "Unable to break out current time\n");

	if( strftime(buffer, 128, "%d-%b-%Y %H:%M:%S", &breakout) == 0 )
		fprintf(stderr, "Unable to convert current time\n");

	printf("%7s: %s: user_time = %ld, system_time = %ld, jiffies = %ld\n",
			message, buffer,
			(long)cur_time.tms_utime, (long)cur_time.tms_stime,
			(long)*the_jiffies);
	fflush(stdout);
	}


/*	this is the stuff that will be traced if user doesn't supply anything else*/
pid_t do_some_stuff( void )
	{
	pid_t		kidpid;
	int			i, j, sum;
	struct tms	tmsstart, tmsstop;
	clock_t		start, stop;
	double		elapsed;

	fflush(stdout);
	if( (kidpid = fork()) < 0 )
		{
		perror("fork");
		sleep(4);
		}
	else if( kidpid == 0 )
		{/* this is the kid */
		printf("job with pid %d started\n", (int)getpid());
		fflush(stdout);
		if( (start = times(&tmsstart)) < 0 )
			exit(EXIT_FAILURE);
		for( i = 0;  i < 4;  i++ )
			{
			sum = 0;
			for( j = 0;  j < 1000000;  j++ )
				sum += j;
			printf("new kid: i = %d, j = %d\n", i, j);
			fflush(stdout);
			sleep(1);
			}
		if( (stop = times(&tmsstop)) < 0 )
			exit(EXIT_FAILURE);
		printf("job with pid %d finished\n", (int)getpid());
		fflush(stdout);
		elapsed = pr_times(stdout, stop-start, &tmsstart, &tmsstop, 1);
		exit(EXIT_SUCCESS);
		}
	/*	this is the parent, just return the kid's pid */
	kpid = kidpid;
	return kidpid;
	}


/* this creates a job requested by user and traces that */
pid_t spawn_stuff( int argc, char *argv[] )
	{
	pid_t	kidpid;

	fflush(stdout);
	if( (kidpid = fork()) < 0 )
		{
		perror("fork");
		}
	else if( kidpid == 0 )
		{/* this is the kid */
		execvp(argv[optind], &argv[optind]);
		perror(argv[optind]);
		exit(EXIT_FAILURE);
		}
	/*	this is the parent, just return the kid's pid */
	kpid = kidpid;
	return kidpid;
	}

/* handler for SIGINT */
void int_handler(int signo)
	{
	
	  printf("\n PARENT HANDLER");
	  fflush(stdout);
	  if(kill(kpid, SIGINT) == -1)
		{
		   perror("\nError in kill");
		}
	}

int main( int argc, char *argv[] )
	{
	pid_t				kidpid, donepid = 0;
	int					status = 0;
	int					c, i, ncpus;
	char				*outfile, *ptr;
	unsigned int		keymask;
	struct sigaction  	sig_act1;
	double				mhz[MAXCPUS];

	sig_act1.sa_handler = int_handler;
	sigemptyset(&sig_act1.sa_mask);
	sig_act1.sa_flags = 0;
	sigaddset(&sig_act1.sa_mask, SIGINT);

	if(sigaction(SIGINT, &sig_act1, NULL))
		{
		  perror("\nsigaction SIGINT\n");
		}

	outfile = NULL;
	keymask = -1;
	opterr = 0;

	while( (c = getopt(argc, argv, OPTIONS)) != EOF )
		switch( c )
			{
		case 'k':
			keymask = strtol(optarg, &ptr, 0);
			if( strspn(ptr, WHITE_SPACE) != strlen(ptr) )
				{
				fprintf(stderr, "illegal keymask parameter %s\n", optarg);
				exit(EXIT_FAILURE);
				}
			break;
		case 'f':
			outfile = optarg;
			break;
		case ':':
			fprintf(stderr, "missing parameter to switch %c\n", optopt);
			exit(EXIT_FAILURE);
		case '?':
			fprintf(stderr, "illegal switch %c\n", optopt);
			exit(EXIT_FAILURE);
			}	/* switch */

	/*	get the cpu info here */
	ncpus = get_cpu_info(MAXCPUS, mhz);
	printf("%7s: %d\n", "n_cpus", ncpus);
	for( i = 0;  i < ncpus;  i++ )
		printf("  cpu %d: %.3f MHZ\n", i, mhz[i]);


	pid = (unsigned long)getpid();
	printf("%7s: %lu\n", "pid", pid);

	/*	force all I/O to be synchronized at this point */
	for( i = 0;  i <= fd;  i++ )
		fsync(i);

	dumptime("setup", &start_time, &start_jiffies);
	errno = 0;
	while( (max_nints = fkt_setup(keymask)) < 0  &&  errno == EAGAIN )
		errno = 0;
	if( max_nints < 0 )
		{
		perror("fkt_setup");
		exit(EXIT_FAILURE);
		}


	if( optind < argc )
		kidpid = spawn_stuff(argc, argv);
	else
		kidpid = do_some_stuff();

	if( kidpid > 0 )
		{/* control came back with kidpid>0, child started ok, wait for it */
		donepid = waitpid(kidpid, &status, 0);
		}

	errno = 0;
	fkt_keychange(FKT_DISABLE, keymask);
	if( errno != 0 )
		perror("fkt_keychange");

	dumptime("stop", &stop_time, &stop_jiffies);

	/* print out stuff about the kid */
	kpid = kidpid;
	printf("job with pid %d started\n", (int)kidpid);
	if( donepid < 0 )
		perror("waitpid");
	else if( WIFEXITED(status) )
		printf("job with pid %d exited with status %d\n",
				(int)kidpid, WEXITSTATUS(status));
	else if( WIFSIGNALED(status) )
		printf("job with pid %d killed by signal %d\n",
				(int)kidpid, WTERMSIG(status));
	else
		printf("job with pid %d terminated for unknown reason\n",
				(int)kidpid);
	fflush(stdout);

	if( outfile == NULL )
		{
		if( (outfile = getenv("TRACE_FILE")) == NULL )
			outfile = DEFAULT_TRACE_FILE;
		}
	while( (fd = open(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0666)) < 0 )
		{/* could not open the indicated file name for writing */
		perror(outfile);
		if( strcmp(outfile, DEFAULT_TRACE_FILE) == 0 )
			{/* can't open the default file, have to give up */
			fkt_endup();
			exit(EXIT_FAILURE);
			}
		outfile = DEFAULT_TRACE_FILE;
		fprintf(stderr, "output defaulting to %s\n", outfile);
		}

	recordit(ncpus, mhz);

	printf("max_nints = %d\n", max_nints);

	errno = 0;
	while( (i = fkt_endup()) != 0  &&  errno == EAGAIN )
		errno = 0;

	if( i != 0 )
		{
		perror("fkt_endup");
		exit(EXIT_FAILURE);
		}

	return 0;
	}
