/*	fkt_setup.c

	Copyright (C) 2000, 2001  Robert D. Russell -- rdr@unh.edu

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/*****	07-Nov-00	rdr		modified for limit rather than amount left *****/

/*	fkt = Fast Kernel Tracing */

#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/fkt.h>

/*	active masks should never become negative numbers in order to prevent
	problems returning them as function results (they look like error codes!) */
#define FULL_ACTIVE_MASK	0x7fffffff

/*	maximum number of regions to allocate when searching for buffer space */
/*	each region has BUFFER_SIZE (32768) integers = 131072 bytes */
#define	MAX_REGIONS	64

/*	3 external ints */

/*	set to non-zero when probing is active */
volatile unsigned int fkt_active = 0;

/*	points to next unused byte in buffer (multiple of 4) */
volatile unsigned int *fkt_next_slot = NULL;

/*	points to slot just beyond end of buffer */
volatile unsigned int *fkt_limit_slot = NULL;


#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)

/*	points to region of allocated buffer space */
static char	*bufptr=NULL;

/*	number of contiguous allocation blocks pointed to by bufptr */
static unsigned int	nblocks = 0;

#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */


/*	called once to set up tracing.
	includes mallocing the buffers to hold the trace.
	returns number of bytes allocated if all ok, else a negative error code.
*/
asmlinkage int sys_fkt_setup( unsigned int keymask )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	unsigned int	base, i, k, n, maxbase, maxk;
	char			*temp, *newptr, *ptr[MAX_REGIONS];
	unsigned long	flags;


	save_flags(flags);
	cli();

	/* paranoia, so nobody waits for us while we are in here */
	fkt_active = 0;

	if( bufptr != NULL )
		{/* previous allocation region was not released, do it now */
		for( n = 0;  n < nblocks;  n++ )
			{
			free_pages((unsigned long)bufptr, FKT_BUFFER_ORDER);
			bufptr += FKT_BUFFER_SIZE*4;
			}
		/* nothing allocated now */
		bufptr = NULL;
		nblocks = 0;
		}

	/*	allocate maximum number of regions of blocks and order them in list */
	for( n = 0;  n < MAX_REGIONS; n++ )
		{
		if( (newptr = (char *)
					__get_free_pages(GFP_ATOMIC, FKT_BUFFER_ORDER)) == NULL )
			break;

		/*****
		printk("allocate region %p, next contiguous would be %p\n",
										newptr, newptr + FKT_BUFFER_SIZE*4);
		*****/

		/*	find slot where this new pointer goes in current list */
		for( i = 0;  i < n  &&  newptr > ptr[i];  i++ )
			;
		/*	shuffle rest of list up to make room for insertion */
		while( i < n )
			{
			temp = ptr[i];
			ptr[i] = newptr;
			newptr = temp;
			i++;
			}
		/*	insert new last item in list */
		ptr[n] = newptr;
		}

	/*	if loop finishes with no regions, we return no memory error */
	if( n == 0 )
		{
		restore_flags(flags);
		return -ENOMEM;
		}

	/*	now find the largest contiguous region of blocks and use that */
	k = 1;								/* 1 block in current region so far */
	base = 0;							/* first region starts here */
	maxk = 1;							/* 1 block in biggest region so far */
	maxbase = 0;						/* first region is biggest so far */
	for( i = 1;  i < n;  i++ )			/* continue search at 2nd region */
		{
		if( ptr[base] + FKT_BUFFER_SIZE*4*k == ptr[i] )
			{/* this block extends current contiguous region from base */
			k++;
			}
		else
			{/* this block starts a new region */
			/*****
			printk("region at %p has %d contiguous regions\n", ptr[base], k);
			*****/
			if( k > maxk )
				{/* ending region was largest so far */
				maxk = k;				/* remember its size */
				maxbase = base;			/* and where it started */
				}
			base = i;					/* new region starts here */
			k = 1;						/* new region contains 1 block so far */
			}
		}
	/*****
	printk("region at %p has %d contiguous regions\n", ptr[base], k);
	*****/
	if( k > maxk )
		{/* last region was the biggest */
		maxk = k;
		maxbase = base;
		}
		
	bufptr = ptr[maxbase];				/* biggest region started here */
	nblocks = maxk;						/* it had this many allocation blocks */

	printk("region at %p MAX %u contiguous regions\n", bufptr, nblocks);

	/*	free up the blocks that precede the region we are keeping */
	for( i = 0;  i < maxbase;  i++ )
		free_pages((unsigned long)ptr[i], FKT_BUFFER_ORDER);

	/*	free up the blocks that follow the region we are keeping */
	for( i = maxbase+maxk;  i < n;  i++ )
		free_pages((unsigned long)ptr[i], FKT_BUFFER_ORDER);

	fkt_next_slot = (unsigned int *)bufptr;
	fkt_limit_slot = (unsigned int *)(bufptr + FKT_BUFFER_SIZE*4*maxk);
	fkt_active = keymask & FULL_ACTIVE_MASK;
	restore_flags(flags);

	if( fkt_active )
		{/* something is being recorded, do the initial probes */
		FKT_PROBE3(-1, FKT_SETUP_CODE, keymask, jiffies, FKT_BUFFER_SIZE*maxk);
		FKT_PROBE0(-1, FKT_CALIBRATE0_CODE);
		FKT_PROBE0(-1, FKT_CALIBRATE0_CODE);
		FKT_PROBE0(-1, FKT_CALIBRATE0_CODE);
		FKT_PROBE1(-1, FKT_CALIBRATE1_CODE, 1);
		FKT_PROBE1(-1, FKT_CALIBRATE1_CODE, 1);
		FKT_PROBE1(-1, FKT_CALIBRATE1_CODE, 1);
		FKT_PROBE2(-1, FKT_CALIBRATE2_CODE, 1, 2);
		FKT_PROBE2(-1, FKT_CALIBRATE2_CODE, 1, 2);
		FKT_PROBE2(-1, FKT_CALIBRATE2_CODE, 1, 2);
		}

	return FKT_BUFFER_SIZE*maxk;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}


/*	called repeatedly to restart tracing.
	returns previous value of fkt_active if all ok, else a negative error code.
*/
asmlinkage int sys_fkt_keychange( int how, unsigned int keymask )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	unsigned int	old_active = fkt_active;


	if( fkt_limit_slot == NULL  ||  bufptr == NULL )
		return -EPERM;

	switch( how )
		{
	case FKT_ENABLE:
		fkt_active |= keymask & FULL_ACTIVE_MASK;
		break;
	case FKT_DISABLE:
		fkt_active &= (~keymask) & FULL_ACTIVE_MASK;
		break;
	case FKT_SETMASK:
		fkt_active = keymask & FULL_ACTIVE_MASK;
		break;
	default:
		return -EINVAL;
		}

	FKT_PROBE2(-1, FKT_KEYCHANGE_CODE, fkt_active, jiffies);
	return old_active;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}


/*	called once to end up tracking.
	includes freeing the buffers.
	returns previous value of active if all ok, else a negative error code.
*/
asmlinkage int sys_fkt_endup( void )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	unsigned int	n;
	unsigned long	flags;
	unsigned int	old_active = fkt_active;


	save_flags(flags);
	cli();

	fkt_active = 0;
	fkt_limit_slot = NULL;
	if( bufptr != NULL )
		{/* previous allocation region was not released, do it now */
		for( n = 0;  n < nblocks;  n++ )
			{
			free_pages((unsigned long)bufptr, FKT_BUFFER_ORDER);
			bufptr += FKT_BUFFER_SIZE*4;
			}
		/* nothing allocated now */
		bufptr = NULL;
		nblocks = 0;
		}

	restore_flags(flags);
	return old_active;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}


/*	called to reset tracing to refill the entire buffer again.
	returns number of bytes in buffer if all ok, else a negative error code.
*/
asmlinkage int sys_fkt_reset( unsigned int keymask )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	unsigned long	flags;
	unsigned int	maxk;


	save_flags(flags);
	cli();

	/* paranoia, so nobody waits for us while we are in here */
	fkt_active = 0;

	if( bufptr == NULL  ||  (maxk = nblocks) == 0 )
		{/* buffer was never allocated, return error */
		restore_flags(flags);
		return -ENOMEM;
		}

	/* reset the buffer to completely empty */
	fkt_next_slot = (unsigned int *)bufptr;
	fkt_limit_slot = (unsigned int *)(bufptr + FKT_BUFFER_SIZE*4*maxk);
	fkt_active = keymask & FULL_ACTIVE_MASK;
	restore_flags(flags);

	if( fkt_active )
		{/* something is being recorded, do the initial probes */
		FKT_PROBE3(-1, FKT_RESET_CODE, keymask, jiffies, FKT_BUFFER_SIZE*maxk);
		FKT_PROBE0(-1, FKT_CALIBRATE0_CODE);
		FKT_PROBE0(-1, FKT_CALIBRATE0_CODE);
		FKT_PROBE0(-1, FKT_CALIBRATE0_CODE);
		FKT_PROBE1(-1, FKT_CALIBRATE1_CODE, 1);
		FKT_PROBE1(-1, FKT_CALIBRATE1_CODE, 1);
		FKT_PROBE1(-1, FKT_CALIBRATE1_CODE, 1);
		FKT_PROBE2(-1, FKT_CALIBRATE2_CODE, 1, 2);
		FKT_PROBE2(-1, FKT_CALIBRATE2_CODE, 1, 2);
		FKT_PROBE2(-1, FKT_CALIBRATE2_CODE, 1, 2);
		}

	return FKT_BUFFER_SIZE*maxk;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}


/*	called repeatedly to copy current buffer into user space.
	returns nints >= 0 if all ok, else a negative error code.
*/
asmlinkage int sys_fkt_getcopy( int maxints, unsigned int *buffer )
	{
	int				nints;
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	int				local_nints = 0;
	int				flag;
	unsigned long	flags;
	char			*local_bufptr;


	save_flags(flags);
	cli();

	/* paranoia, so nobody waits for us while we are in here */
	fkt_active = 0;

	if( bufptr == NULL )
		{
		nints = -EPERM;
		goto out;
		}

	/*	get number of ints worth of trace data currently in the buffer */
	local_nints = nints = ((char *)fkt_next_slot - bufptr)/4;

	if( maxints == 0 )
		{/*	user just wants to know how many ints are in the buffer */
		goto out;
		}

	if( maxints < nints )
		{/*	user did not give us enough space to store entire buffer */
		nints = -EINVAL;
		goto out;
		}

	/*	first ensure that any buffer supplied by the user is writable */
	if( buffer )
		{/* do the memory accessing outside our mutex lock */
		local_bufptr = bufptr;
		restore_flags(flags);
		if( (flag = verify_area(VERIFY_WRITE, buffer, maxints*4)) )
			{/* user gave us some bad memory, don't use it */
			nints = flag;
			}
		else
			{/*	buffer writable, fill it in with the trace information */
			copy_to_user(buffer, local_bufptr, nints*4);
			}
		return nints;
		}
	else
		{/* user gave us a NULL pointer to memory */
		nints = -EINVAL;
		}

out:
	restore_flags(flags);
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	nints = -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return nints;
	}


asmlinkage int sys_fkt_probe0( unsigned int keymask, unsigned int code )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	if( keymask & fkt_active )
		fkt_header((code<<8) | 12);
	return 0;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}


asmlinkage int sys_fkt_probe1( unsigned int keymask, unsigned int code,
															unsigned int p1 )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	if( keymask & fkt_active )
		fkt_header((code<<8) | 16, p1);
	return 0;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}


asmlinkage int sys_fkt_probe2( unsigned int keymask, unsigned int code,
											unsigned int p1, unsigned int p2 )
	{
#if defined(CONFIG_FKT)  ||  defined(CONFIG_FKT_TIME_ONLY)
	if( keymask & fkt_active )
		fkt_header((code<<8) | 20, p1, p2);
	return 0;
#else	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	return -ENOSYS;
#endif	/* CONFIG_FKT or CONFIG_FKT_TIME_ONLY */
	}
